# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
import eventlet

from neutron_lib.db import api as db_api
from neutron_lib.services import base as service_base
from oslo_log import log
from oslo_serialization import jsonutils

from networking_bigswitch.plugins.bigswitch.db import consistency_db
from networking_bigswitch.plugins.bigswitch.db import network_template_db
from networking_bigswitch.plugins.bigswitch.db import reachability_test_db
from networking_bigswitch.plugins.bigswitch.db import tenant_policy_db
from networking_bigswitch.plugins.bigswitch.extensions \
    import bsnserviceextension
from networking_bigswitch.plugins.bigswitch import servermanager

LOG = log.getLogger(__name__)


class BSNServicePlugin(service_base.ServicePluginBase,
                       bsnserviceextension.BSNServicePluginBase):

    supported_extension_aliases = ["bsn-service-extension"]

    def __init__(self):
        super(BSNServicePlugin, self).__init__()
        # initialize BCF server handler
        self.servers = servermanager.ServerPool.get_instance()
        self.networktemplate_db_mixin = network_template_db\
            .NetworkTemplateDbMixin()
        self.network_template_assignment_db_mixin = network_template_db\
            .NetworkTemplateAssignmentDbMixin()
        self.reachabilitytest_db_mixin = reachability_test_db\
            .ReachabilityTestDbMixin()
        self.reachabilityquicktest_db_mixin = reachability_test_db\
            .ReachabilityQuickTestDbMixin()
        self.tenantpolicy_db_mixin = tenant_policy_db.TenantPolicyDbMixin()

    def get_plugin_type(self):
        # Tell Neutron this is a BSN service plugin
        return 'BSNSERVICEPLUGIN'

    def get_plugin_name(self):
        return 'bsn_service_extension'

    def get_plugin_description(self):
        return "Arista CCF Service Plugin"

    # public CRUD methods for network templates
    def get_networktemplates(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        return self.networktemplate_db_mixin.get_networktemplates(
            context=context, filters=filters, fields=fields, sorts=sorts,
            limit=limit, marker=marker, page_reverse=page_reverse)

    def get_networktemplate(self, context, id, fields=None):
        return self.networktemplate_db_mixin.get_networktemplate(
            context=context, id=id, fields=fields)

    def create_networktemplate(self, context, networktemplate):
        return self.networktemplate_db_mixin.create_networktemplate(
            context=context, networktemplate=networktemplate)

    def delete_networktemplate(self, context, id):
        self.networktemplate_db_mixin.delete_networktemplate(
            context=context, id=id)

    def update_networktemplate(self, context, id, networktemplate):
        return self.networktemplate_db_mixin.update_networktemplate(
            context=context, id=id, networktemplate=networktemplate)

    # public CRUD methods for Network Template Assignment
    def get_networktemplateassignments(self, context, filters=None,
                                       fields=None, sorts=None, limit=None,
                                       marker=None, page_reverse=False):
        return self.network_template_assignment_db_mixin\
            .get_networktemplateassignments(
                context=context, filters=filters, fields=fields, sorts=sorts,
                limit=limit, marker=marker, page_reverse=page_reverse)

    def get_networktemplateassignment(self, context, id, fields=None):
        return self.network_template_assignment_db_mixin\
            .get_networktemplateassignment(context=context, id=id,
                                           fields=fields)

    def create_networktemplateassignment(self, context,
                                         networktemplateassignment):
        return self.network_template_assignment_db_mixin\
            .create_networktemplateassignment(
                context=context,
                networktemplateassignment=networktemplateassignment)

    def delete_networktemplateassignment(self, context, id):
        self.network_template_assignment_db_mixin\
            .delete_networktemplateassignment(context=context, id=id)

    def update_networktemplateassignment(self, context, id,
                                         networktemplateassignment):
        return self.network_template_assignment_db_mixin\
            .update_networktemplateassignment(
                context=context, id=id,
                networktemplateassignment=networktemplateassignment)

    # common method to parse response from the controller
    def parse_result(self, response, expected_result):
        if not response:
            test_result = "fail"
            detail = [{'path-index': "No result", 'hop-index': '',
                       'hop-name': ''}]
            logical_path = [{'path-index': "No result", 'hop-index': '',
                             'hop-name': ''}]
            return test_result, detail, logical_path
        elif response[0].get("summary", [{}])[0].get("forward-result") != \
                expected_result:
            test_result = "fail"
            detail = [{'path-index': "Expected: %s. Actual: %s" % (
                expected_result,
                response[0].get("summary", [{}])[0].get("forward-result"))}]
            try:
                detail[0]['path-index'] += \
                    " - " + response[0]['summary'][0]['logical-error']
            except Exception:
                pass
        elif response[0].get("summary", [{}])[0].get("forward-result") == \
                expected_result:
            test_result = "pass"
            detail = response[0].get("physical-path", [{}])
        else:
            try:
                detail = [{'path-index':
                           response[0]['summary'][0]['logical-error']}]
            except Exception:
                detail = [{'path-index': jsonutils.dumps(response)}]
            test_result = "fail"
        detail[0]['path-index'] = detail[0].get('path-index', '')
        detail[0]['hop-index'] = detail[0].get('hop-index', '')
        detail[0]['hop-name'] = detail[0].get('hop-name', '')

        # also get logical-path
        logical_path = response[0].get("logical-path", [{}])
        return test_result, detail, logical_path

    # public CRUD methods for Reachability Test
    def get_reachabilitytests(self, context, filters=None, fields=None,
                              sorts=None, limit=None, marker=None,
                              page_reverse=False):
        return self.reachabilitytest_db_mixin.get_reachabilitytests(
            context, filters=filters, fields=fields, sorts=sorts, limit=limit,
            marker=marker, page_reverse=page_reverse)

    def get_reachabilitytest(self, context, id, fields=None):
        return self.reachabilitytest_db_mixin.get_reachabilitytest(
            context=context, id=id, fields=fields)

    def create_reachabilitytest(self, context, reachabilitytest):
        return self.reachabilitytest_db_mixin.create_reachabilitytest(
            context=context, reachabilitytest=reachabilitytest)

    def update_reachabilitytest(self, context, id, reachabilitytest):
        reachabilitytest_data = reachabilitytest['reachabilitytest']
        reachabilitytest = self.reachabilitytest_db_mixin\
            ._get_reachabilitytest(context, id)
        if 'run_test' in reachabilitytest_data and \
                reachabilitytest_data['run_test']:
            # run test on the controller and get results
            # update fields in reachabilitytest_data dict
            src = reachabilitytest.get_connection_source(
                unicode_mode=self.servers.is_unicode_enabled())
            dst = reachabilitytest.get_connection_destination()
            response = self.servers.rest_get_testpath(src, dst)
            test_result, detail, logical_path = self.parse_result(
                response, reachabilitytest.expected_result)
            reachabilitytest_data['test_result'] = test_result
            reachabilitytest_data['detail'] = detail
            reachabilitytest_data['logical_path'] = logical_path
            # reset run_test to false and set timestamp to now
            reachabilitytest_data['run_test'] = False
            reachabilitytest_data['test_time'] = datetime.now()
        return self.reachabilitytest_db_mixin.update_reachabilitytest(
            context=context, id=id,
            reachabilitytest={'reachabilitytest': reachabilitytest_data})

    def delete_reachabilitytest(self, context, id):
        self.reachabilitytest_db_mixin.delete_reachabilitytest(
            context=context, id=id)

    # public CRUD methods for Reachability Quick Test
    def get_reachabilityquicktests(self, context, filters=None, fields=None,
                                   sorts=None, limit=None, marker=None,
                                   page_reverse=False):
        return self.reachabilityquicktest_db_mixin.get_reachabilityquicktests(
            context=context, filters=filters, fields=fields, sorts=sorts,
            limit=limit, marker=marker, page_reverse=page_reverse)

    def get_reachabilityquicktest(self, context, id, fields=None):
        return self.reachabilityquicktest_db_mixin.get_reachabilityquicktest(
            context=context, id=id, fields=fields)

    def create_reachabilityquicktest(self, context, reachabilityquicktest):
        return self.reachabilityquicktest_db_mixin\
            .create_reachabilityquicktest(
                context=context, reachabilityquicktest=reachabilityquicktest)

    def update_reachabilityquicktest(self, context, id, reachabilityquicktest):
        reachabilityquicktest_data = \
            reachabilityquicktest['reachabilityquicktest']
        reachabilityquicktest = self.reachabilityquicktest_db_mixin\
            ._get_reachabilityquicktest(context, id)
        if 'save_test' in reachabilityquicktest_data and \
                reachabilityquicktest_data['save_test']:
            # just copy paste the test and return
            quicktest = self.reachabilityquicktest_db_mixin\
                ._get_reachabilityquicktest(context, id)
            # update name as given in the args, not the default name
            quicktest.name = reachabilityquicktest_data['name']
            # remove ID, as we want it to be unique per test
            # not unique per tenant
            quicktest_dict = self.reachabilityquicktest_db_mixin\
                ._make_reachabilityquicktest_dict(quicktest)
            quicktest_dict.pop('id')
            self.reachabilitytest_db_mixin.create_reachabilitytest_withresult(
                context=context,
                reachabilitytest={'reachabilitytest': quicktest_dict})
            # reset the save_test flag
            reachabilityquicktest_data['save_test'] = False
        if 'run_test' in reachabilityquicktest_data and \
                reachabilityquicktest_data['run_test']:
            # run test on the controller and get results
            # update fields in reachabilityquicktest_data dict
            src = reachabilityquicktest.get_connection_source(
                unicode_mode=self.servers.is_unicode_enabled())
            dst = reachabilityquicktest.get_connection_destination()
            response = self.servers.rest_get_testpath(src, dst)
            test_result, detail, logical_path = self.parse_result(
                response, reachabilityquicktest.expected_result)
            reachabilityquicktest_data['test_result'] = test_result
            reachabilityquicktest_data['detail'] = detail
            reachabilityquicktest_data['logical_path'] = logical_path
            # reset run_test to false and set timestamp to now
            reachabilityquicktest_data['run_test'] = False
            reachabilityquicktest_data['test_time'] = datetime.now()
        return self.reachabilityquicktest_db_mixin\
            .update_reachabilityquicktest(
                context=context, id=id,
                reachabilityquicktest={'reachabilityquicktest':
                                       reachabilityquicktest_data})

    def delete_reachabilityquicktest(self, context, id):
        self.reachabilityquicktest_db_mixin.delete_reachabilityquicktest(
            context=context, id=id)

    # public CRUD methods for Tenant Policies
    def get_tenantpolicies(self, context, filters=None, fields=None,
                           sorts=None, limit=None, marker=None,
                           page_reverse=False):
        return self.tenantpolicy_db_mixin.get_tenantpolicies(
            context=context, filters=filters, fields=fields, sorts=sorts,
            limit=limit, marker=marker, page_reverse=page_reverse)

    def get_tenantpolicy(self, context, id, fields=None):
        return self.tenantpolicy_db_mixin.get_tenantpolicy(
            context=context, id=id, fields=fields)

    def create_tenantpolicy(self, context, tenantpolicy):
        with db_api.CONTEXT_WRITER.using(context):
            tenantpolicy_dict = self.tenantpolicy_db_mixin.create_tenantpolicy(
                context=context, tenantpolicy=tenantpolicy)
            self.servers.rest_create_tenantpolicy(
                tenantpolicy_dict['tenant_id'], tenantpolicy_dict)

            return tenantpolicy_dict

    def delete_tenantpolicy(self, context, id):
        with db_api.CONTEXT_WRITER.using(context):
            delete_policy = self.tenantpolicy_db_mixin._get_tenantpolicy(
                context, id)
            self.tenantpolicy_db_mixin.delete_tenantpolicy(
                context=context, id=id)
            self.servers.rest_delete_tenantpolicy(delete_policy['tenant_id'],
                                                  delete_policy['priority'])

    def update_tenantpolicy(self, context, id, tenantpolicy):
        with db_api.CONTEXT_WRITER.using(context):
            updated_policy = self.tenantpolicy_db_mixin.update_tenantpolicy(
                context=context, servers=self.servers, id=id,
                tenantpolicy=tenantpolicy)
            self.servers.rest_update_tenantpolicy(updated_policy['tenant_id'],
                                                  updated_policy)

            return updated_policy

    # public CRUD methods for Topology Sync command
    def update_forcesynctopology(self, context, id, forcesynctopology):
        eventlet.spawn(self.servers.force_topo_sync, **{'check_ts': False})
        return {'id': '1',
                'tenant_id': context.project_id,
                'project_id': context.project_id,
                'timestamp_ms': '0',
                'timestamp_datetime': '0',
                'status': 'Topology Sync scheduled for execution.'}

    def get_forcesynctopologies(self, context, filters=None, fields=None,
                                sorts=None, limit=None, marker=None,
                                page_reverse=False):
        with context.session.begin(subtransactions=True):
            res = (context.session.query(consistency_db.ConsistencyHash).
                   filter_by(hash_id='1').first())
            if res:
                if 'TOPO_SYNC' in res.hash:
                    timestamp_ms = consistency_db.get_lock_owner(res.hash)
                    timestamp_datetime = consistency_db.convert_ts_to_datetime(
                        timestamp_ms)
                    result = 'Topology sync in progress..'
                else:
                    timestamp_ms = res.hash
                    timestamp_datetime = consistency_db.convert_ts_to_datetime(
                        timestamp_ms)
                    result = 'Topology sync complete.'
                # return the result
                return [{'id': '1',
                         'tenant_id': context.project_id,
                         'project_id': context.project_id,
                         'timestamp_ms': timestamp_ms,
                         'timestamp_datetime': timestamp_datetime,
                         'status': result}]
            else:
                return [{'id': '1',
                         'tenant_id': context.project_id,
                         'project_id': context.project_id,
                         'timestamp_ms': '0',
                         'timestamp_datetime': '0',
                         'status': 'FAILURE'}]

    def get_forcesynctopology(self, context, id, fields=None):
        return self.get_forcesynctopologies(context=context)[0]



# Copyright 2017 Big Switch Networks, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslotest import base


class TestImports(base.BaseTestCase):

    # This test simply import files that are not used in other UT.
    # the reasoning for this is that if any upstream dependencies
    # are broken, it is cheapest to find out here.
    def test_imports(self):
        import networking_bigswitch.plugins.bigswitch.bsn_service_plugin  # noqa
        import networking_bigswitch.plugins.bigswitch.extensions.bsnserviceextension  # noqa



# Copyright 2017, Big Switch Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add tenant/segment id for reachability tests

Revision ID: 938c3e0e2029
Revises:
Create Date: 2018-10-15 11:04:57.644914

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '938c3e0e2029'
down_revision = '7db8cd315b95'
branch_labels = None
depends_on = None


def upgrade():
    # reachability test
    # segment name can be empty for uuid based schema (supported queens/5.0)
    op.alter_column(
        table_name='reachabilitytest',
        column_name='src_segment_name',
        existing_type=sa.String(255, convert_unicode=True),
        nullable=True)

    # If unicode is off, tenant id can be empty
    op.add_column('reachabilitytest',
                  sa.Column('src_tenant_id', sa.String(36), nullable=True))

    # If unicode is off, segment id can be empty
    op.add_column('reachabilitytest',
                  sa.Column('src_segment_id', sa.String(36), nullable=True))

    # reachability quick test
    # segment name can be empty for uuid based schema (supported queens/5.0)
    op.alter_column(
        table_name='reachabilityquicktest',
        column_name='src_segment_name',
        existing_type=sa.String(255),
        nullable=True)

    # If unicode is off, tenant id can be empty
    op.add_column('reachabilityquicktest',
                  sa.Column('src_tenant_id', sa.String(36), nullable=True))

    # If unicode is off, segment id can be empty
    op.add_column('reachabilityquicktest',
                  sa.Column('src_segment_id', sa.String(36), nullable=True))


def downgrade():
    pass



# Copyright 2012 Big Switch Networks, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Neutron REST Proxy Plug-in for Big Switch and FloodLight Controllers.

NeutronRestProxy provides a generic neutron plugin that translates all plugin
function calls to equivalent authenticated REST calls to a set of redundant
external network controllers. It also keeps persistent store for all neutron
state to allow for re-sync of the external controller(s), if required.

The local state on the plugin also allows for local response and fast-fail
semantics where it can be determined based on the local persistent store.

Network controller specific code is decoupled from this plugin and expected
to reside on the controller itself (via the REST interface).

This allows for:
 - independent authentication and redundancy schemes between neutron and the
   network controller
 - independent upgrade/development cycles between neutron and the controller
   as it limits the proxy code upgrade requirement to neutron release cycle
   and the controller specific code upgrade requirement to controller code
 - ability to sync the controller with neutron for independent recovery/reset

External REST API used by proxy is the same API as defined for neutron (JSON
subset) with some additional parameters (gateway on network-create and macaddr
on port-attach) on an additional PUT to do a bulk dump of all persistent data.
"""

import copy
import functools
import re

import eventlet
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_utils import importutils
from six.moves import http_client
from sqlalchemy.orm import exc as sqlexc

from neutron.agent import rpc as agent_rpc
from neutron.agent import securitygroups_rpc as sg_rpc
from neutron.api import extensions as neutron_extensions
from neutron.api.rpc.agentnotifiers import dhcp_rpc_agent_api
from neutron.api.rpc.handlers import dhcp_rpc
from neutron.api.rpc.handlers import metadata_rpc
from neutron.api.rpc.handlers import securitygroups_rpc
from neutron.db import agents_db
from neutron.db import agentschedulers_db
from neutron.db import allowedaddresspairs_db as addr_pair_db
from neutron.db import db_base_plugin_v2
from neutron.db import external_net_db
from neutron.db import extradhcpopt_db
from neutron.db import l3_db
from neutron.db.models import securitygroup as sg_db
from neutron.db import models_v2
from neutron.db import securitygroups_rpc_base as sg_db_rpc

from neutron_lib.agent import topics
from neutron_lib.api.definitions import allowedaddresspairs as addr_pair
from neutron_lib.api.definitions import external_net as extnet_apidef
from neutron_lib.api.definitions import extra_dhcp_opt as edo_ext
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as const
from neutron_lib import context as qcontext
from neutron_lib.db import api as db_api
from neutron_lib.db import utils as db_utils
from neutron_lib import exceptions as lib_exceptions
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from neutron_lib.utils import runtime

from networking_bigswitch.plugins.bigswitch import config as pl_config
from networking_bigswitch.plugins.bigswitch import constants as bsn_constants
from networking_bigswitch.plugins.bigswitch.db import porttracker_db
from networking_bigswitch.plugins.bigswitch import extensions
from networking_bigswitch.plugins.bigswitch.i18n import _
from networking_bigswitch.plugins.bigswitch.i18n import _LW
from networking_bigswitch.plugins.bigswitch import servermanager
from networking_bigswitch.plugins.bigswitch.utils import Util
from networking_bigswitch.plugins.bigswitch import version

LOG = logging.getLogger(__name__)

SYNTAX_ERROR_MESSAGE = _('Syntax error in server config file, aborting plugin')
METADATA_SERVER_IP = '169.254.169.254'


class AgentNotifierApi(securitygroups_rpc.SecurityGroupAgentRpcApiMixin):

    def __init__(self, topic):
        self.topic = topic
        target = oslo_messaging.Target(topic=topic, version='1.0')
        self.client = n_rpc.get_client(target)

    def port_update(self, context, port):
        topic_port_update = topics.get_topic_name(self.client.target.topic,
                                                  topics.PORT, topics.UPDATE)
        cctxt = self.client.prepare(fanout=True, topic=topic_port_update)
        cctxt.cast(context, 'port_update', port=port)


class SecurityGroupServerRpcMixin(sg_db_rpc.SecurityGroupServerRpcMixin):

    def get_port_from_device(self, context, device):
        port_id = re.sub(r"^%s" % const.TAP_DEVICE_PREFIX, "", device)
        port = self.get_port_and_sgs(context, port_id)
        if port:
            port['device'] = device
        return port

    def get_port_and_sgs(self, context, port_id):
        """Get port from database with security group info."""

        LOG.debug("get_port_and_sgs() called for port_id %s", port_id)
        sg_binding_port = sg_db.SecurityGroupPortBinding.port_id

        with db_api.CONTEXT_READER.using(context):
            query = context.session.query(
                models_v2.Port,
                sg_db.SecurityGroupPortBinding.security_group_id
            )
            query = query.outerjoin(sg_db.SecurityGroupPortBinding,
                                    models_v2.Port.id == sg_binding_port)
            query = query.filter(models_v2.Port.id.startswith(port_id))
            port_and_sgs = query.all()
            if not port_and_sgs:
                return
            port = port_and_sgs[0][0]
            plugin = directory.get_plugin()
            port_dict = plugin._make_port_dict(port)
            port_dict['security_groups'] = [
                sg_id for port_, sg_id in port_and_sgs if sg_id]
            port_dict['security_group_rules'] = []
            port_dict['security_group_source_groups'] = []
            port_dict['fixed_ips'] = [ip['ip_address']
                                      for ip in port['fixed_ips']]
        return port_dict


class NeutronRestProxyV2Base(db_base_plugin_v2.NeutronDbPluginV2,
                             external_net_db.External_net_db_mixin,
                             SecurityGroupServerRpcMixin,
                             agentschedulers_db.DhcpAgentSchedulerDbMixin):

    supported_extension_aliases = ["binding"]
    servers = None

    def __init__(self):
        super(NeutronRestProxyV2Base, self).__init__()
        self._setup_rpc()

    def _setup_rpc(self):
        self.conn = n_rpc.Connection()
        self.topic = topics.PLUGIN
        self.notifier = AgentNotifierApi(topics.AGENT)
        # init dhcp agent support
        self._dhcp_agent_notifier = dhcp_rpc_agent_api.DhcpAgentNotifyAPI()
        self.agent_notifiers[const.AGENT_TYPE_DHCP] = (
            self._dhcp_agent_notifier
        )
        self.endpoints = [agent_rpc.PluginApi(self.topic),
                          securitygroups_rpc.SecurityGroupServerRpcCallback(),
                          dhcp_rpc.DhcpRpcCallback(),
                          agents_db.AgentExtRpcCallback(),
                          metadata_rpc.MetadataRpcCallback()]
        self.conn.create_consumer(self.topic, self.endpoints,
                                  fanout=False)
        # Consume from all consumers in threads
        self.conn.consume_in_threads()

    @property
    def l3_plugin(self):
        return directory.get_plugin(plugin_constants.L3)

    @property
    def bsn_service_plugin(self):
        return directory.get_plugin(bsn_constants.BSN_SERVICE_PLUGIN)

    def _validate_names(self, obj, name=None):
        """validate names

        :returns
            True, if obj name, obj's tenant name and name have supported chars
            False, otherwise
        """

        if self.servers.is_unicode_enabled():
            return True

        if name and not servermanager.is_valid_bcf_name(name):
            LOG.warning('Unsupported characters in Name: %(name)s. ',
                        {'name': name})
            return False

        if (obj and 'name' in obj and
                not servermanager.is_valid_bcf_name(obj['name'])):
            LOG.warning('Unsupported characters in Name: %(name)s. '
                        'Object details: %(obj)s',
                        {'name': obj['name'], 'obj': obj})
            return False

        if (obj and 'tenant_name' in obj and
                not servermanager.is_valid_bcf_name(obj['tenant_name'])):
            LOG.warning('Unsupported characters in TenantName: %(tname)s. '
                        'Object details: %(obj)s',
                        {'tname': obj['tenant_name'], 'obj': obj})
            return False

        return True

    def _get_all_data_auto(self):
        return self._get_all_data(
            get_sgs=cfg.CONF.RESTPROXY.sync_security_groups)

    def _get_all_data(self, get_ports=True, get_sgs=True):
        # sync tenant cache with keystone
        if not self.servers._update_tenant_cache(reconcile=False):
            return None

        admin_context = qcontext.get_admin_context()
        networks = []
        # this method is used by the ML2 driver so it can't directly invoke
        # the self.get_(ports|networks) methods
        plugin = directory.get_plugin()
        all_networks = plugin.get_networks(admin_context) or []
        for net in all_networks:
            try:
                if self._skip_bcf_network_event(net):
                    LOG.info('Skipping segment create for Network: %(n)s',
                             {'n': net.get('name')})
                    continue

                mapped_network = self._get_mapped_network_with_subnets(net)
                if not self._validate_names(mapped_network):
                    continue
                # validate names for subnet as well
                if 'subnets' in mapped_network:
                    new_subnets = []
                    for subnet in mapped_network['subnets']:
                        if not self._validate_names(subnet):
                            continue
                        new_subnets.append(subnet)
                    mapped_network['subnets'] = new_subnets

                networks_n_ports = mapped_network

                if get_ports:
                    ports = []
                    net_filter = {'network_id': [net.get('id')]}
                    net_ports = plugin.get_ports(admin_context,
                                                 filters=net_filter) or []
                    for port in net_ports:
                        if not self._is_port_supported(port):
                            continue

                        mapped_port = self._map_display_name_or_tenant(port)
                        if self.servers.is_unicode_enabled():
                            # remove port name so that it won't be stored in
                            #  description
                            mapped_port['name'] = None
                        mapped_port = self._map_state_and_status(mapped_port)
                        mapped_port = self._map_port_hostid(mapped_port, net)
                        if not mapped_port:
                            continue

                        mapped_port['attachment'] = {
                            'id': port.get('device_id'),
                            'mac': port.get('mac_address'),
                        }

                        if not get_sgs:
                            if 'security_groups' in mapped_port:
                                del mapped_port['security_groups']

                        ports.append(mapped_port)

                    networks_n_ports['ports'] = ports

                if networks_n_ports:
                    networks.append(networks_n_ports)
            except servermanager.TenantIDNotFound:
                # if tenant name is not known to keystone, skip the network
                continue

        data = {'networks': networks}

        if get_sgs and self.l3_plugin:
            sgs = plugin.get_security_groups(admin_context) or []
            new_sgs = []
            for sg in sgs:
                try:
                    mapped_sg = self._map_display_name_or_tenant(sg)
                    if not self._validate_names(mapped_sg):
                        continue
                    if 'description' in mapped_sg:
                        mapped_sg['description'] = ''
                    if self.servers.is_unicode_enabled():
                        mapped_sg['name'] = None
                    else:
                        mapped_sg['name'] = Util.format_resource_name(
                            mapped_sg['name'])
                    new_sgs.append(mapped_sg)
                except servermanager.TenantIDNotFound:
                    # if tenant name is not known to keystone, skip the sg
                    continue

            data.update({'security-groups': new_sgs})

        all_tenants_map = self.servers.keystone_tenants

        if self.servers.is_unicode_enabled():
            # display-name is only supported as list for topology in NSAPI
            tenants = []
            for tenant_id, tenant_name in all_tenants_map.items():
                tenants.append({
                    'name': tenant_id,
                    'id': tenant_id,
                    'display-name': tenant_name
                })
        else:
            # dict for tenant works in topology sync only if display-name is
            # not enabled
            tenants = {}
            for tenant in all_tenants_map:
                if not self._validate_names(None,
                                            name=all_tenants_map[tenant]):
                    continue
                tenants[tenant] = all_tenants_map[tenant]

        data['tenants'] = tenants
        return data

    def _send_all_data_auto(self, timeout=None, triggered_by_tenant=None):
        return self._send_all_data(
            timeout=timeout,
            triggered_by_tenant=triggered_by_tenant)

    def _send_all_data(self, send_ports=True,
                       send_sgs=True, timeout=None,
                       triggered_by_tenant=None):
        """Pushes all data to network ctrl (networks/ports, ports/attachments).

        This gives the controller an option to re-sync it's persistent store
        with neutron's current view of that data.

        All args are ignored. The `_get_all_data` method dynamically pulls the
        relevant information i.e. starting from train, this supports l2 only
        """
        sync_executed, topo_resp = self.servers.force_topo_sync()
        return topo_resp

    def _assign_resource_to_service_tenant(self, resource):
        resource['tenant_id'] = (resource['tenant_id'] or
                                 servermanager.SERVICE_TENANT)

        if not self.servers.is_unicode_enabled():
            if resource.get('name'):
                # resource name may contain space. Replace space with -
                resource['name'] = Util.format_resource_name(resource['name'])

    def _get_all_subnets_json_for_network(self, net_id, context=None):
        if context is None:
            context = qcontext.get_admin_context()
        # start a sub-transaction to avoid breaking parent transactions
        with db_api.CONTEXT_WRITER.using(context):
            subnets = self._get_subnets_by_network(context,
                                                   net_id)
        subnets_details = []
        if subnets:
            for subnet in subnets:
                subnet_dict = self._make_subnet_dict(subnet, context=context)
                mapped_subnet = self._map_display_name_or_tenant(subnet_dict)
                mapped_subnet = self._map_state_and_status(mapped_subnet)
                subnets_details.append(mapped_subnet)

        return subnets_details

    def _tenant_check_for_security_group(self, sg):
        """Router VRRP creates a hidden network for router heart-beats.

        This network is not associated with any tenant
        """
        sg['tenant_id'] = sg['tenant_id'] or servermanager.SERVICE_TENANT
        tenant_name = self.servers.keystone_tenants.get(sg['tenant_id'])

        if not tenant_name:
            self.servers._update_tenant_cache(reconcile=True)
            tenant_name = self.servers.keystone_tenants.get(sg['tenant_id'])

        if not self.servers.is_unicode_enabled():
            sg['tenant_name'] = tenant_name
        return tenant_name

    def bsn_create_security_group(self, sg_id=None, sg=None, context=None):
        if sg_id:
            if not hasattr(context, 'session'):
                context = qcontext.get_admin_context()
            # overwrite sg if both sg and sg_id are given
            sg = self.get_security_group(context, sg_id)

        if sg:
            if self.servers.is_unicode_enabled():
                sg['display-name'] = sg['name']
                sg['name'] = None
            else:
                sg['name'] = Util.format_resource_name(sg['name'])
            # remove description as its not used
            if sg.get('description'):
                del(sg['description'])
            # check and map tenant_name for sg
            tenant_name = self._tenant_check_for_security_group(sg)
            # skip the security group if its tenant is unknown
            if tenant_name:
                if tenant_name == servermanager.SERVICE_TENANT:
                    self.bsn_create_tenant(servermanager.SERVICE_TENANT,
                                           context=context)
                self.servers.rest_create_securitygroup(sg)
        else:
            LOG.warning(_LW("No security group is provided for creation."))

    def bsn_delete_security_group(self, sg_id, context=None):
        self.servers.rest_delete_securitygroup(sg_id)

    def bsn_create_tenant(self, tenant_id, context=None):
        self.servers.rest_create_tenant(tenant_id)

    def bsn_delete_tenant(self, tenant_id, context=None):
        self.servers.rest_delete_tenant(tenant_id)

    def _verify_network_precommit(self, context):
        if not self.servers.is_unicode_enabled():
            if context.current['name'] != context.original['name']:
                raise servermanager.NetworkNameChangeError()

    def _get_mapped_network_with_subnets(self, network, context=None):
        # if context is not provided, admin context is used
        if context is None:
            context = qcontext.get_admin_context()
        network = self._map_display_name_or_tenant(network)
        network = self._map_state_and_status(network)
        subnets = self._get_all_subnets_json_for_network(network['id'],
                                                         context)
        network['subnets'] = subnets
        for subnet in (subnets or []):
            if subnet['gateway_ip']:
                # FIX: For backward compatibility with wire protocol
                network['gateway'] = subnet['gateway_ip']
                break
        else:
            network['gateway'] = ''
        network[extnet_apidef.EXTERNAL] = self._network_is_external(
            context, network['id'])
        # include ML2 segmentation types
        network['segmentation_types'] = getattr(self, "segmentation_types", "")
        # OSP-45: remove name to avoid NSAPI error in convertToAscii
        for subnet in (subnets or []):
            subnet.pop('name', None)

        return network

    def _skip_bcf_network_event(self, network):
        '''Check if the network event needs to be sent to BCF

        return true, if event should be skipped, i.e. not be sent to BCF
               false, otherwise
       '''
        pnet = network.get(pl_config.PROVIDER_PHYSNET)
        if pnet and pl_config.SRIOV_ACTIVE_ACTIVE_MODE_PHYSNET_SUBSTR in pnet:
            # Configure BCF segment only for networks on ACTIVE physnet
            if not pnet.endswith(pl_config.SRIOV_ACTIVE_PHYSNET):
                return True
        return False

    def _send_create_network(self, network, context=None):
        tenant_id = network['tenant_id']
        if self._skip_bcf_network_event(network):
            LOG.info('Skipping CCF segment create for Network: %(name)s',
                     {'name': network.get('name')})
            return

        if context is None:
            context = qcontext.get_admin_context()
        filters = {'name': ['default'], 'tenant_id': [tenant_id]}
        if cfg.CONF.RESTPROXY.sync_security_groups:
            default_group = self.get_security_groups(
                context, filters, default_sg=True)
            if default_group:
                # VRRP tenant doesn't have tenant_id
                self.bsn_create_security_group(sg=default_group[0])
        # display-name is also mapped here
        mapped_network = self._get_mapped_network_with_subnets(network,
                                                               context)

        if not tenant_id:
            tenant_id = servermanager.SERVICE_TENANT
            mapped_network['tenant_id'] = servermanager.SERVICE_TENANT
            if not self.servers.is_unicode_enabled():
                mapped_network['name'] = Util.format_resource_name(
                    mapped_network['name'])
            self.bsn_create_tenant(servermanager.SERVICE_TENANT,
                                   context=context)
        self.servers.rest_create_network(tenant_id, mapped_network)

    def _send_update_network(self, network, context=None):
        net_id = network['id']
        tenant_id = network['tenant_id']
        if self._skip_bcf_network_event(network):
            LOG.info('Skipping CCF segment update for Network: %(name)s',
                     {'name': network.get('name')})
            return

        # display-name is also mapped here
        mapped_network = self._get_mapped_network_with_subnets(network,
                                                               context)

        if not tenant_id:
            tenant_id = servermanager.SERVICE_TENANT
            mapped_network['tenant_id'] = servermanager.SERVICE_TENANT
            if not self.servers.is_unicode_enabled():
                mapped_network['name'] = Util.format_resource_name(
                    mapped_network['name'])
        self.servers.rest_update_network(tenant_id, net_id, mapped_network)

    def _send_delete_network(self, network, context=None):
        net_id = network['id']
        tenant_id = network['tenant_id'] or servermanager.SERVICE_TENANT
        self.servers.rest_delete_network(tenant_id, net_id)

    def _map_display_name_or_tenant(self, resource):
        """This maps tenant_name or display-name for an object

        None-unicode mode uses tenant_name
        Unicode mode uses tenant_id and display-name

        :param resource: object to be mapped
        :return: mapped object copy
        """
        resource = copy.deepcopy(resource)
        self._assign_resource_to_service_tenant(resource)

        tenant_name = self.servers.keystone_tenants.get(resource['tenant_id'])
        if not tenant_name:
            self.servers._update_tenant_cache()
            tenant_name = self.servers.keystone_tenants.get(
                resource['tenant_id'])
            if not tenant_name:
                raise servermanager.TenantIDNotFound(
                    tenant=resource['tenant_id'])

        if self.servers.is_unicode_enabled():
            if resource.get('name'):
                resource['display-name'] = resource['name']
            # cases like network needs the name on bcf side
            resource['name'] = resource['id']
        else:
            resource['tenant_name'] = tenant_name

        return resource

    def _map_state_and_status(self, resource):
        resource = copy.copy(resource)
        resource['state'] = ('UP' if resource.pop('admin_state_up',
                                                  True) else 'DOWN')
        resource.pop('status', None)

        return resource

    def _is_port_supported(self, port):
        """Check if the vnic-type is supported

        :return: True, if port is supported
                 False, otherwise
        """
        vnic_type = port.get(portbindings.VNIC_TYPE)
        if vnic_type and vnic_type in pl_config.UNSUPPORTED_VNIC_TYPES:
            return False
        return True

    def _is_port_sriov(self, port):
        """Check if port is an SR-IOV port

        :return: True, if port is an SR-IOV port
                 False, otherwise
        """
        vnic_type = port.get(portbindings.VNIC_TYPE)
        if vnic_type and vnic_type in pl_config.VNIC_TYPE_SRIOV:
            return True
        return False

    def _sriov_port_validation_active_active(self, port, network):
        """For SR-IOV port, we configure 'memeber interface-group $HOSTID' on BCF.

        In Active-Active mode, this is done ONLY for ports belonging to the
        ACTIVE physnet. In Active-Backup mode, this is done for all ports.

        :param port:
        :param network:
        :return: boolean value whether it passed or failed
        """
        network_type = network.get(pl_config.PROVIDER_NETWORK_TYPE)
        if not network_type or network_type != const.TYPE_VLAN:
            return False

        physnet = network.get(pl_config.PROVIDER_PHYSNET)
        if not physnet:
            return False

        if pl_config.SRIOV_ACTIVE_ACTIVE_MODE_PHYSNET_SUBSTR in physnet:
            # Active-Active mode, configure BCF only for ACTIVE physnet
            if not physnet.endswith(pl_config.SRIOV_ACTIVE_PHYSNET):
                return False

        return True

    def _is_port_sriov_vm_detach(self, port, network):
        """We allow empty host_id field during update_port_postcommit.

        This is a check to see if the port is SRIOV port and a case of VM
        detach. If yes, we send a delete_port to the BCF controller for that
        port. Otherwise return false and do nothing for the port.

        :param port:
        :param network:
        :return: boolean value specifying if port is sriov_vm_detach case
        """
        # if port is SRIOV and unbound, it is VM detach
        if self._is_port_sriov(port):
            if not self._sriov_port_validation_active_active(port, network):
                return False

            vif_type = port.get(portbindings.VIF_TYPE)
            if not vif_type:
                return False
            elif vif_type == portbindings.VIF_TYPE_UNBOUND:
                return True
        return False

    def _get_sriov_port_hostid(self, port, network):
        """Return the HostID for the given SR-IOV port

        For SR-IOV port, we configure 'memeber interface-group $HOSTID' on BCF.
        In Active-Active mode, this is done ONLY for ports belonging to the
        ACTIVE physnet. In Active-Backup mode, this is done for all ports.
        HostID = $(H)-$(PHYSNET), which corresponds to interface-groups on BCF.

        :return: HostID, if membership-rule needs to be configured on BCF
                 None, otherwise
        """
        if not self._sriov_port_validation_active_active(port, network):
            return None
        physnet = network.get(pl_config.PROVIDER_PHYSNET)
        bsn_host_id = port.get(portbindings.HOST_ID) + '-' + physnet
        return bsn_host_id

    def _get_ovs_dpdk_port_hostid(self, port, network):
        """Return HostID with bridge_name appended for OVS and DPDK ports

        VIF type for OVS and DPDK ports is OVS and VHOSTUSER respectively.

        :param port:
        :param network:
        :return: new_host_id
        """
        host_id = port.get(portbindings.HOST_ID)
        physnet = network.get(pl_config.PROVIDER_PHYSNET)
        if physnet not in self.bridge_mappings:
            LOG.warning(_LW("Physical network to bridge mapping not "
                            "found for port %s."), port)
            return host_id

        bridge_name = self.bridge_mappings.get(physnet)
        host_id_bridge_name = host_id + '_' + bridge_name
        return host_id_bridge_name

    def _map_port_hostid(self, port, network):
        """Update the HOST_ID of a given port based on it's type.

        Perform basic sanity checks and update the HOST_ID of the port
        :return: port, if port is of relevance to BCF
                 False, otherwise
        """
        prepped_port = copy.copy(port)
        if (portbindings.HOST_ID not in prepped_port or
                prepped_port[portbindings.HOST_ID] == ''):
            LOG.debug("Ignoring port notification to controller because of "
                      "missing host ID.")
            return False

        # Update HOST_ID (to be used by BCF).
        # - For SR-IOV it is a function of HostID & physnet info
        if self._is_port_sriov(prepped_port):
            vif_type = prepped_port.get(portbindings.VIF_TYPE)
            if not vif_type or vif_type == portbindings.VIF_TYPE_UNBOUND:
                # Port not bound yet, nothing to do
                return False

            hostid = self._get_sriov_port_hostid(prepped_port, network)
            if not hostid:
                return False
            prepped_port[portbindings.HOST_ID] = hostid

        # update HOST_ID to '<host-id>_<bridge-name>' for ports with
        # VIF_TYPE OVS and VHOSTUSER i.e. DHCP and DPDK ports
        # if bridge_name not available, sets it to just 'host-id'
        vif_type = prepped_port.get(portbindings.VIF_TYPE)
        if (vif_type and
                (vif_type == portbindings.VIF_TYPE_OVS or
                 vif_type == portbindings.VIF_TYPE_VHOST_USER)):
            prepped_port[portbindings.HOST_ID] = (
                self._get_ovs_dpdk_port_hostid(prepped_port, network))

        return prepped_port

    def _warn_on_state_status(self, resource):
        if resource.get('admin_state_up', True) is False:
            LOG.warning("Setting admin_state_up=False is not supported "
                        "in this plugin version. Ignoring setting for "
                        "resource: %s", resource)

        if 'status' in resource:
            if resource['status'] != const.NET_STATUS_ACTIVE:
                LOG.warning("Operational status is internally set by the "
                            "plugin. Ignoring setting status=%s.",
                            resource['status'])

    def _get_router_intf_details(self, context, port, subnet_id):
        """This gets router interface for 1 single subnet

        For multiple subnets, this should be called multiple times
        :param context:
        :param port:
        :param subnet_id:
        :return:
        """
        # we will use the network id as interface's id
        subnet = self.get_subnet(context, subnet_id)
        net_id = subnet['network_id']
        network = self.get_network(context, net_id)
        mapped_network = self._get_mapped_network_with_subnets(network)
        mapped_subnet = self._map_display_name_or_tenant(subnet)
        mapped_subnet = self._map_state_and_status(mapped_subnet)

        data = {
            'id': subnet_id,
            "network": mapped_network,
            "subnet": mapped_subnet
        }

        # get gateway_ip from port instead of gateway_ip
        # every port can have multiple fixed_ips
        if port.get("fixed_ips"):
            for fixed_ip in port["fixed_ips"]:
                if fixed_ip['subnet_id'] == subnet_id:
                    data['ip_address'] = fixed_ip['ip_address']
                    break

        return data

    def _extend_port_dict_binding(self, context, port):
        cfg_vif_type = cfg.CONF.NOVA.vif_type.lower()
        if cfg_vif_type not in (portbindings.VIF_TYPE_OVS):
            LOG.warning("Unrecognized vif_type in configuration "
                        "[%s]. Defaulting to ovs.",
                        cfg_vif_type)
            cfg_vif_type = portbindings.VIF_TYPE_OVS
        # In ML2, the host_id is already populated
        if portbindings.HOST_ID in port:
            hostid = port[portbindings.HOST_ID]
        elif 'id' in port:
            hostid = porttracker_db.get_port_hostid(context, port['id'])
        else:
            hostid = None
        if hostid:
            port[portbindings.HOST_ID] = hostid
            override = self._check_hostvif_override(hostid)
            if override:
                cfg_vif_type = override
        port[portbindings.VIF_TYPE] = cfg_vif_type

        sg_enabled = sg_rpc.is_firewall_enabled()
        port[portbindings.VIF_DETAILS] = {
            # TODO(rkukura): Replace with new VIF security details
            portbindings.CAP_PORT_FILTER:
            'security-group' in self.supported_extension_aliases,
            portbindings.OVS_HYBRID_PLUG: sg_enabled
        }
        return port

    def _check_hostvif_override(self, hostid):
        for v in cfg.CONF.NOVA.vif_types:
            if hostid in getattr(cfg.CONF.NOVA, "node_override_vif_" + v, []):
                return v
        return False

    def _get_port_net_tenantid(self, context, port):
        net = super(NeutronRestProxyV2Base,
                    self).get_network(context, port["network_id"])
        return net['tenant_id']

    def _add_service_tenant_to_port(self, port):
        port['tenant_id'] = port['tenant_id'] or servermanager.SERVICE_TENANT

        if 'network' in port:
            port['network']['tenant_id'] = (
                port['network']['tenant_id'] or servermanager.SERVICE_TENANT)

    def async_port_create(self, tenant_id, net_id, port, update_status=True):
        try:
            tenant_id = tenant_id or servermanager.SERVICE_TENANT
            rest_port = copy.deepcopy(port)
            self._add_service_tenant_to_port(rest_port)
            self.servers.rest_create_port(tenant_id, net_id, rest_port)
        except servermanager.RemoteRestError as e:
            # 404 should never be received on a port create unless
            # there are inconsistencies between the data in neutron
            # and the data in the backend.
            # Run a sync to get it consistent.
            if (cfg.CONF.RESTPROXY.auto_sync_on_failure and
                    e.status == http_client.NOT_FOUND and
                    servermanager.NXNETWORK in e.reason):
                LOG.error("Inconsistency with backend controller "
                          "triggering full synchronization.")
                # args depend on if we are operating in ML2 driver
                # or as the full plugin
                self._send_all_data_auto(triggered_by_tenant=tenant_id)
                # If the full sync worked, the port will be created
                # on the controller so it can be safely marked as active
            else:
                # Any errors that don't result in a successful auto-sync
                # require that the port be placed into the error state.
                LOG.error(
                    "NeutronRestProxyV2: Unable to create port: %s", e)
                try:
                    self._set_port_status(port['id'], const.PORT_STATUS_ERROR)
                except lib_exceptions.PortNotFound:
                    # If port is already gone from DB and there was an error
                    # creating on the backend, everything is already consistent
                    pass
                return

        if not update_status:
            return

        new_status = (const.PORT_STATUS_ACTIVE if port['state'] == 'UP'
                      else const.PORT_STATUS_DOWN)
        try:
            self._set_port_status(port['id'], new_status)
        except lib_exceptions.PortNotFound:
            # This port was deleted before the create made it to the controller
            # so it now needs to be deleted since the normal delete request
            # would have deleted an non-existent port.
            tenant_id = tenant_id or servermanager.SERVICE_TENANT
            self.servers.rest_delete_port(tenant_id, net_id, port['id'])

    # NOTE(kevinbenton): workaround for eventlet/mysql deadlock
    @runtime.synchronized('bsn-port-barrier')
    def _set_port_status(self, port_id, status):
        session = db_api.get_writer_session()
        try:
            port = session.query(models_v2.Port).filter_by(id=port_id).one()
            port['status'] = status
            session.flush()
        except sqlexc.NoResultFound:
            raise lib_exceptions.PortNotFound(port_id=port_id)


def add_debug_log(f):
    @functools.wraps(f)
    def wrapper(self, context, *args, **kwargs):
        # core plugin: context is top level object
        # Do not need context in header anymore
        LOG.debug("Function: %(fname)s called", {'fname': f.__name__})
        return f(self, context, *args, **kwargs)
    return wrapper


class NeutronRestProxyV2(NeutronRestProxyV2Base,
                         addr_pair_db.AllowedAddressPairsMixin,
                         extradhcpopt_db.ExtraDhcpOptMixin):

    _supported_extension_aliases = ["external-net", "binding",
                                    "extra_dhcp_opt", "quotas",
                                    "dhcp_agent_scheduler", "agent",
                                    "security-group", "allowed-address-pairs"]

    @property
    def supported_extension_aliases(self):
        if not hasattr(self, '_aliases'):
            aliases = self._supported_extension_aliases[:]
            sg_rpc.disable_security_group_extension_by_config(aliases)
            self._aliases = aliases
        return self._aliases

    def __init__(self):
        super(NeutronRestProxyV2, self).__init__()
        LOG.info('NeutronRestProxy: Starting plugin. Version=%s',
                 version.version_string_with_vcs())
        pl_config.register_config()
        self.evpool = eventlet.GreenPool(cfg.CONF.RESTPROXY.thread_pool_size)

        # Include the Big Switch Extensions path in the api_extensions
        neutron_extensions.append_api_extensions_path(extensions.__path__)

        self.add_meta_server_route = cfg.CONF.RESTPROXY.add_meta_server_route

        # init network ctrl connections
        self.servers = servermanager.ServerPool()
        self.servers.get_topo_function = self._get_all_data_auto

        self.network_scheduler = importutils.import_object(
            cfg.CONF.network_scheduler_driver
        )

        # setup rpc for security and DHCP agents
        self._setup_rpc()

        if cfg.CONF.RESTPROXY.sync_data:
            self._send_all_data_auto()

        self.add_periodic_dhcp_agent_status_check()
        LOG.debug("NeutronRestProxyV2: initialization done")

    @db_api.CONTEXT_WRITER
    @add_debug_log
    def create_network(self, context, network):
        """Create a network.

        Network represents an L2 network segment which can have a set of
        subnets and ports associated with it.

        :param context: neutron api request context
        :param network: dictionary describing the network

        :returns: a sequence of mappings with the following signature:
        {
            "id": UUID representing the network.
            "name": Human-readable name identifying the network.
            "tenant_id": Owner of network. NOTE: only admin user can specify
                         a tenant_id other than its own.
            "admin_state_up": Sets admin state of network.
                              if down, network does not forward packets.
            "status": Indicates whether network is currently operational
                      (values are "ACTIVE", "DOWN", "BUILD", and "ERROR")
            "subnets": Subnets associated with this network.
        }

        :raises: RemoteRestError
        """
        self._warn_on_state_status(network['network'])

        self._ensure_default_security_group(
            context,
            network['network']["tenant_id"]
        )
        # create network in DB
        new_net = super(NeutronRestProxyV2, self).create_network(context,
                                                                 network)
        self._process_l3_create(context, new_net, network['network'])
        # create network on the network controller
        self._send_create_network(new_net, context)

        # return created network
        return new_net

    @db_api.CONTEXT_WRITER
    @add_debug_log
    def update_network(self, context, net_id, network):
        """Updates the properties of a particular Virtual Network.

        :param context: neutron api request context
        :param net_id: uuid of the network to update
        :param network: dictionary describing the updates

        :returns: a sequence of mappings with the following signature:
        {
            "id": UUID representing the network.
            "name": Human-readable name identifying the network.
            "tenant_id": Owner of network. NOTE: only admin user can
                         specify a tenant_id other than its own.
            "admin_state_up": Sets admin state of network.
                              if down, network does not forward packets.
            "status": Indicates whether network is currently operational
                      (values are "ACTIVE", "DOWN", "BUILD", and "ERROR")
            "subnets": Subnets associated with this network.
        }

        :raises: exceptions.NetworkNotFound
        :raises: RemoteRestError
        """
        self._warn_on_state_status(network['network'])

        new_net = super(NeutronRestProxyV2, self).update_network(
            context, net_id, network)
        self._process_l3_update(context, new_net, network['network'])

        # update network on network controller
        self._send_update_network(new_net, context)
        return new_net

    @db_api.CONTEXT_WRITER
    @add_debug_log
    def delete_network(self, context, net_id):
        """Delete a network.

        :param context: neutron api request context
        :param id: UUID representing the network to delete.

        :returns: None

        :raises: exceptions.NetworkInUse
        :raises: exceptions.NetworkNotFound
        :raises: RemoteRestError
        """
        # Validate args
        orig_net = super(NeutronRestProxyV2, self).get_network(context, net_id)

        self._process_l3_delete(context, net_id)
        ret_val = super(NeutronRestProxyV2, self).delete_network(context,
                                                                 net_id)
        self._send_delete_network(orig_net, context)
        return ret_val

    @add_debug_log
    def create_port(self, context, port):
        """Create a port.

        create a port, which is a connection point of a device
        (e.g., a VM NIC) to attach an L2 Neutron network.
        :param context: neutron api request context
        :param port: dictionary describing the port

        :returns:
        {
            "id": uuid representing the port.
            "network_id": uuid of network.
            "tenant_id": tenant_id
            "mac_address": mac address to use on this port.
            "admin_state_up": Sets admin state of port. if down, port
                              does not forward packets.
            "status": dicates whether port is currently operational
                      (limit values to "ACTIVE", "DOWN", "BUILD", and "ERROR")
            "fixed_ips": list of subnet IDs and IP addresses to be used on
                         this port
            "device_id": identifies the device (e.g., virtual server) using
                         this port.
        }

        :raises: exceptions.NetworkNotFound
        :raises: exceptions.StateInvalid
        :raises: RemoteRestError
        """
        # Update DB in new session so exceptions rollback changes
        with db_api.CONTEXT_WRITER.using(context):
            self._ensure_default_security_group_on_port(context, port)
            sgids = self._get_security_groups_on_port(context, port)
            # non-router port status is set to pending. it is then updated
            # after the async rest call completes. router ports are synchronous
            if port['port']['device_owner'] == l3_db.DEVICE_OWNER_ROUTER_INTF:
                port['port']['status'] = const.PORT_STATUS_ACTIVE
            elif not port['port'].get('status'):
                port['port']['status'] = const.PORT_STATUS_BUILD
            dhcp_opts = port['port'].get(edo_ext.EXTRADHCPOPTS, [])
            new_port = super(NeutronRestProxyV2, self).create_port(context,
                                                                   port)
            self._process_port_create_security_group(context, new_port, sgids)
        if (portbindings.HOST_ID in port['port'] and 'id' in new_port):
            host_id = port['port'][portbindings.HOST_ID]
            porttracker_db.put_port_hostid(context, new_port['id'],
                                           host_id)
        new_port[addr_pair.ADDRESS_PAIRS] = (
            self._process_create_allowed_address_pairs(
                context, new_port,
                port['port'].get(addr_pair.ADDRESS_PAIRS)))
        self._process_port_create_extra_dhcp_opts(context, new_port,
                                                  dhcp_opts)
        new_port = self._extend_port_dict_binding(context, new_port)
        net = super(NeutronRestProxyV2,
                    self).get_network(context, new_port["network_id"])
        if self.add_meta_server_route:
            if new_port['device_owner'] == const.DEVICE_OWNER_DHCP:
                destination = METADATA_SERVER_IP + '/32'
                self._add_host_route(context, destination, new_port)

        # create on network ctrl
        mapped_port = self._map_display_name_or_tenant(new_port)
        if self.servers.is_unicode_enabled():
            # remove port name so that it won't be stored in description
            mapped_port['name'] = None
        mapped_port = self._map_state_and_status(mapped_port)
        # ports have to be created synchronously when creating a router
        # port since adding router interfaces is a multi-call process
        if mapped_port['device_owner'] == l3_db.DEVICE_OWNER_ROUTER_INTF:
            self.servers.rest_create_port(net["tenant_id"],
                                          new_port["network_id"],
                                          mapped_port)
        else:
            self.evpool.spawn_n(self.async_port_create, net["tenant_id"],
                                new_port["network_id"], mapped_port)
        self.notify_security_groups_member_updated(context, new_port)
        return new_port

    def get_port(self, context, id, fields=None):
        with db_api.CONTEXT_READER.using(context):
            port = super(NeutronRestProxyV2, self).get_port(context, id,
                                                            fields)
            self._extend_port_dict_binding(context, port)
        return db_utils.resource_fields(port, fields)

    def get_ports(self, context, filters=None, fields=None):
        with db_api.CONTEXT_READER.using(context):
            ports = super(NeutronRestProxyV2, self).get_ports(context, filters,
                                                              fields)
            for port in ports:
                self._extend_port_dict_binding(context, port)
        return [db_utils.resource_fields(port, fields) for port in ports]

    @add_debug_log
    def update_port(self, context, port_id, port):
        """Update values of a port.

        :param context: neutron api request context
        :param id: UUID representing the port to update.
        :param port: dictionary with keys indicating fields to update.

        :returns: a mapping sequence with the following signature:
        {
            "id": uuid representing the port.
            "network_id": uuid of network.
            "tenant_id": tenant_id
            "mac_address": mac address to use on this port.
            "admin_state_up": sets admin state of port. if down, port
                               does not forward packets.
            "status": dicates whether port is currently operational
                       (limit values to "ACTIVE", "DOWN", "BUILD", and "ERROR")
            "fixed_ips": list of subnet IDs and IP addresses to be used on
                         this port
            "device_id": identifies the device (e.g., virtual server) using
                         this port.
        }

        :raises: exceptions.StateInvalid
        :raises: exceptions.PortNotFound
        :raises: RemoteRestError
        """
        self._warn_on_state_status(port['port'])

        # Validate Args
        orig_port = super(NeutronRestProxyV2, self).get_port(context, port_id)
        with db_api.CONTEXT_WRITER.using(context):
            # Update DB
            new_port = super(NeutronRestProxyV2,
                             self).update_port(context, port_id, port)
            ctrl_update_required = False
            if addr_pair.ADDRESS_PAIRS in port['port']:
                ctrl_update_required |= (
                    self.update_address_pairs_on_port(context, port_id, port,
                                                      orig_port, new_port))
            self._update_extra_dhcp_opts_on_port(context, port_id, port,
                                                 new_port)
            old_host_id = porttracker_db.get_port_hostid(context,
                                                         orig_port['id'])
            if (portbindings.HOST_ID in port['port'] and 'id' in new_port):
                host_id = port['port'][portbindings.HOST_ID]
                porttracker_db.put_port_hostid(context, new_port['id'],
                                               host_id)
                if old_host_id != host_id:
                    ctrl_update_required = True

            if (new_port.get("device_id") != orig_port.get("device_id") and
                    orig_port.get("device_id")):
                ctrl_update_required = True

            if ctrl_update_required:
                # tenant_id must come from network in case network is shared
                net_tenant_id = self._get_port_net_tenantid(context, new_port)
                new_port = self._extend_port_dict_binding(context, new_port)
                mapped_port = self._map_display_name_or_tenant(new_port)
                if self.servers.is_unicode_enabled():
                    # remove port name so that it won't be stored in
                    # description
                    mapped_port['name'] = None
                mapped_port = self._map_state_and_status(mapped_port)
                self.servers.rest_update_port(net_tenant_id,
                                              new_port["network_id"],
                                              mapped_port)
            need_port_update_notify = self.update_security_group_on_port(
                context, port_id, port, orig_port, new_port)
        need_port_update_notify |= self.is_security_group_member_updated(
            context, orig_port, new_port)

        if need_port_update_notify:
            self.notifier.port_update(context, new_port)

        # return new_port
        return new_port

    # NOTE(kevinbenton): workaround for eventlet/mysql deadlock
    @runtime.synchronized('bsn-port-barrier')
    @add_debug_log
    def delete_port(self, context, port_id, l3_port_check=True):
        """Delete a port.

        :param context: neutron api request context
        :param id: UUID representing the port to delete.

        :raises: exceptions.PortInUse
        :raises: exceptions.PortNotFound
        :raises: exceptions.NetworkNotFound
        :raises: RemoteRestError
        """
        # if needed, check to see if this is a port owned by
        # an l3-router.  If so, we should prevent deletion.
        if l3_port_check and self.l3_plugin:
            self.l3_plugin.prevent_l3_port_deletion(context, port_id)
        with db_api.CONTEXT_WRITER.using(context):
            if self.l3_plugin:
                router_ids = self.l3_plugin.disassociate_floatingips(
                    context, port_id, do_notify=False)
            port = super(NeutronRestProxyV2, self).get_port(context, port_id)
            # Tenant ID must come from network in case the network is shared
            tenid = self._get_port_net_tenantid(context, port)
            self.ipam.delete_port(context, port_id)
            tenid = tenid or servermanager.SERVICE_TENANT
            self.servers.rest_delete_port(tenid, port['network_id'], port_id)

        if self.l3_plugin:
            # now that we've left db transaction, we are safe to notify
            self.l3_plugin.notify_routers_updated(context, router_ids)

    @add_debug_log
    def create_subnet(self, context, subnet):
        self._warn_on_state_status(subnet['subnet'])

        with context.session.begin(subtransactions=True):
            # create subnet in DB
            new_subnet = super(NeutronRestProxyV2,
                               self).create_subnet(context, subnet)
            net_id = new_subnet['network_id']
            orig_net = super(NeutronRestProxyV2,
                             self).get_network(context, net_id)
            # update network on network controller
            self._send_update_network(orig_net, context)
        return new_subnet

    @add_debug_log
    def update_subnet(self, context, id, subnet):
        self._warn_on_state_status(subnet['subnet'])

        with db_api.CONTEXT_WRITER.using(context):
            # update subnet in DB
            new_subnet = super(NeutronRestProxyV2,
                               self).update_subnet(context, id, subnet)
            net_id = new_subnet['network_id']
            orig_net = super(NeutronRestProxyV2,
                             self).get_network(context, net_id)
            # update network on network controller
            self._send_update_network(orig_net, context)
            return new_subnet

    @add_debug_log
    def delete_subnet(self, context, id):
        orig_subnet = super(NeutronRestProxyV2, self).get_subnet(context, id)
        net_id = orig_subnet['network_id']
        with db_api.CONTEXT_WRITER.using(context):
            # delete subnet in DB
            super(NeutronRestProxyV2, self).delete_subnet(context, id)
            orig_net = super(NeutronRestProxyV2, self).get_network(context,
                                                                   net_id)
            # update network on network controller - exception will rollback
            self._send_update_network(orig_net, context)

    def _add_host_route(self, context, destination, port):
        subnet = {}
        for fixed_ip in port['fixed_ips']:
            subnet_id = fixed_ip['subnet_id']
            nexthop = fixed_ip['ip_address']
            subnet['host_routes'] = [{'destination': destination,
                                      'nexthop': nexthop}]
            updated_subnet = self.update_subnet(context,
                                                subnet_id,
                                                {'subnet': subnet})
            payload = {'subnet': updated_subnet}
            self._dhcp_agent_notifier.notify(context, payload,
                                             'subnet.update.end')
            LOG.debug("Adding host route: ")
            LOG.debug("Destination:%(dst)s nexthop:%(next)s",
                      {'dst': destination, 'next': nexthop})



# Copyright 2012 Big Switch Networks, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from mock import patch
# from neutron.tests.unit import _test_extension_portbindings as test_bindings
from neutron.tests.unit.api.v2 import test_base
from neutron.tests.unit.db import test_allowedaddresspairs_db as test_addr_pair
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_plugin
# from neutron_lib.api.definitions import portbindings
# from neutron_lib import constants
from neutron_lib import context
from neutron_lib.plugins import directory
from oslo_config import cfg
import webob.exc

# from networking_bigswitch.plugins.bigswitch import config as pl_config
from networking_bigswitch.plugins.bigswitch.servermanager import \
    TenantIDNotFound
from networking_bigswitch.tests.unit.bigswitch import fake_server
from networking_bigswitch.tests.unit.bigswitch.mock_paths import HTTPCON
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    IS_UNICODE_ENABLED
from networking_bigswitch.tests.unit.bigswitch \
    import test_base as bsn_test_base


class BigSwitchProxyPluginV2TestCase(bsn_test_base.BigSwitchTestBase,
                                     test_plugin.NeutronDbPluginV2TestCase):

    def setUp(self, plugin_name=None, service_plugins=None, ext_mgr=None):
        if hasattr(self, 'HAS_PORT_FILTER'):
            cfg.CONF.set_override(
                'enable_security_group', self.HAS_PORT_FILTER, 'SECURITYGROUP')
        self.setup_config_files()
        self.setup_patches()
        if plugin_name:
            self._plugin_name = plugin_name

        test_plugin.NeutronDbPluginV2TestCase.setUp(
            self, self._plugin_name,
            service_plugins=service_plugins,
            ext_mgr=ext_mgr)

        self.port_create_status = 'BUILD'
        self.startHttpPatch()

    def setup_coreplugin(self, plugin, load_plugins=False):
        self.setup_db()
        super(BigSwitchProxyPluginV2TestCase, self).setup_coreplugin(
            plugin, load_plugins=load_plugins)


class TestBigSwitchProxyBasicGet(test_plugin.TestBasicGet,
                                 BigSwitchProxyPluginV2TestCase):

    pass


class TestBigSwitchProxyV2HTTPResponse(test_plugin.TestV2HTTPResponse,
                                       BigSwitchProxyPluginV2TestCase):

    def test_failover_memory(self):
        # first request causes failover so next shouldn't hit bad server
        with self.network() as net:
            kwargs = {'tenant_id': 'ExceptOnBadServer'}
            with self.network(**kwargs) as net:
                req = self.new_show_request('networks', net['network']['id'])
                res = req.get_response(self.api)
                self.assertEqual(res.status_int, 200)

# TODO(weifan): This blocks gating, skip for now, needs to be fixed
# functionally, port creation works and tempest tests pass

# class TestBigSwitchProxyPortsV2(test_plugin.TestPortsV2,
#                                 BigSwitchProxyPluginV2TestCase,
#                                 test_bindings.PortBindingsTestCase):
#
#     VIF_TYPE = portbindings.VIF_TYPE_OVS
#     HAS_PORT_FILTER = False
#
#     def setUp(self, plugin_name=None):
#         super(TestBigSwitchProxyPortsV2,
#               self).setUp(self._plugin_name)
#
#     def test_get_ports_no_id(self):
#         with self.port(name='test'):
#             ports = directory.get_plugin().get_ports(
#                 context.get_admin_context(), fields=['name'])
#             self.assertEqual(['name'], list(ports[0].keys()))
#
#     def test_router_port_status_active(self):
#         # router ports screw up port auto-deletion so it has to be
#         # disabled for this test
#         with self.network() as net:
#             with self.subnet(network=net) as sub:
#                 with self.port(
#                     subnet=sub,
#                     do_delete=False,
#                     device_owner=constants.DEVICE_OWNER_ROUTER_INTF
#                 ) as port:
#                     # router ports should be immediately active
#                     self.assertEqual(port['port']['status'], 'ACTIVE')
#
#     def test_update_port_status_build(self):
#         # normal ports go into the pending build state for async creation
#         with self.port() as port:
#             self.assertEqual(port['port']['status'], 'BUILD')
#             self.assertEqual(self.port_create_status, 'BUILD')
#
#     def _get_ports(self, netid):
#         return self.deserialize('json',
#                                 self._list_ports('json',
#                                                  netid=netid))['ports']
#
#     def test_rollback_for_port_create(self):
#         plugin = directory.get_plugin()
#         with self.subnet() as s:
#             # stop normal patch
#             self.httpPatch.stop()
#             # allow thread spawns for this test
#             self.spawn_p.stop()
#             kwargs = {'device_id': 'somedevid'}
#             # put in a broken 'server'
#             httpPatch = patch(HTTPCON, new=fake_server.HTTPConnectionMock500)
#             httpPatch.start()
#             with self.port(subnet=s, **kwargs):
#                 # wait for async port create request to finish
#                 plugin.evpool.waitall()
#                 # put good 'server' back in
#                 httpPatch.stop()
#                 self.httpPatch.start()
#                 ports = self._get_ports(s['subnet']['network_id'])
#                 # failure to create should result in port in error state
#                 self.assertEqual(ports[0]['status'], 'ERROR')
#
#     def test_rollback_for_port_update(self):
#         with self.network() as n:
#             with self.port(network_id=n['network']['id'],
#                            device_id='66') as port:
#                 port = self._get_ports(n['network']['id'])[0]
#                 data = {'port': {'name': 'aNewName', 'device_id': '99'}}
#                 # stop normal patch
#                 self.httpPatch.stop()
#                 with patch(HTTPCON, new=fake_server.HTTPConnectionMock500):
#                     self.new_update_request(
#                         'ports', data, port['id']).get_response(self.api)
#                 self.httpPatch.start()
#                 uport = self._get_ports(n['network']['id'])[0]
#                 # name should have stayed the same
#                 self.assertEqual(port['name'], uport['name'])
#
#     def test_rollback_for_port_delete(self):
#         with self.network() as n:
#             with self.port(network_id=n['network']['id'],
#                            device_id='somedevid') as port:
#                 # stop normal patch
#                 self.httpPatch.stop()
#                 with patch(HTTPCON, new=fake_server.HTTPConnectionMock500):
#                     self._delete(
#                         'ports',
#                         port['port']['id'],
#                         expected_code=webob.exc.HTTPInternalServerError.code)
#                 self.httpPatch.start()
#                 port = self._get_ports(n['network']['id'])[0]
#                 self.assertEqual('BUILD', port['status'])
#
#     def test_correct_shared_net_tenant_id(self):
#         # tenant_id in port requests should match network tenant_id instead
#         # of port tenant_id
#         def rest_port_op(self, ten_id, netid, port):
#             if ten_id != 'SHARED':
#                 raise Exception('expecting tenant_id SHARED. got %s'
#                                 % ten_id)
#         with self.network(tenant_id='SHARED', shared=True) as net:
#             with self.subnet(network=net) as sub:
#                 pref = ('networking_bigswitch.plugins.bigswitch'
#                         '.servermanager.ServerPool.%s')  # noqa
#                 tomock = [pref % 'rest_create_port',
#                           pref % 'rest_update_port',
#                           pref % 'rest_delete_port']
#                 patches = [patch(f, create=True, new=rest_port_op)
#                            for f in tomock]
#                 for restp in patches:
#                     restp.start()
#                 with self.port(subnet=sub, tenant_id='port-owner') as port:
#                     data = {'port': {'binding:host_id': 'someotherhost',
#                             'device_id': 'override_dev'}}
#                     req = self.new_update_request('ports', data,
#                                                   port['port']['id'])
#                     res = req.get_response(self.api)
#                     self.assertEqual(res.status_int, 200)
#
#     def test_create404_triggers_sync(self):
#         # allow async port thread for this patch
#         self.spawn_p.stop()
#         with\
#             self.subnet() as s,\
#             patch(HTTPCON, create=True,
#                   new=fake_server.HTTPConnectionMock404),\
#             patch(bsn_test_base.PLUGIN_PATH +
#                   '.NeutronRestProxyV2._send_all_data') as mock_send_all:
#
#             with self.port(subnet=s, device_id='somedevid') as p:
#                 # wait for the async port thread to finish
#                 plugin = directory.get_plugin()
#                 plugin.evpool.waitall()
#         call = mock.call(
#             send_routers=True, send_floating_ips=True, timeout=None,
#             triggered_by_tenant=p['port']['tenant_id']
#         )
#         mock_send_all.assert_has_calls([call])
#         self.spawn_p.start()
#
#     def test_port_vif_details_default(self):
#         kwargs = {'name': 'name', 'device_id': 'override_dev'}
#         with self.port(**kwargs) as port:
#             self.assertEqual(port['port']['binding:vif_type'],
#                              portbindings.VIF_TYPE_OVS)
#
#     def test_port_vif_details_override(self):
#         # ivshost is in the test config to override to IVS
#         kwargs = {'name': 'name', 'binding:host_id': 'ivshost',
#                   'device_id': 'override_dev',
#                   'arg_list': ('binding:host_id',)}
#         with self.port(**kwargs) as port:
#             self.assertEqual(port['port']['binding:vif_type'],
#                              pl_config.VIF_TYPE_IVS)
#         self._delete('ports', port['port']['id'])
#         self._delete('networks', port['port']['network_id'])
#         kwargs = {'name': 'name2', 'binding:host_id': 'someotherhost',
#                   'device_id': 'other_dev'}
#         with self.port(**kwargs) as port:
#             self.assertEqual(port['port']['binding:vif_type'], self.VIF_TYPE)
#
#     def test_port_move(self):
#         # ivshost is in the test config to override to IVS
#         kwargs = {'name': 'name', 'binding:host_id': 'ivshost',
#                   'device_id': 'override_dev'}
#         with self.port(**kwargs) as port:
#             data = {'port': {'binding:host_id': 'someotherhost',
#                              'device_id': 'override_dev'}}
#             req = self.new_update_request('ports', data, port['port']['id'])
#             res = self.deserialize(self.fmt, req.get_response(self.api))
#             self.assertEqual(res['port']['binding:vif_type'], self.VIF_TYPE)


class TestVifDifferentDefault(BigSwitchProxyPluginV2TestCase):

    def setup_config_files(self):
        super(TestVifDifferentDefault, self).setup_config_files()
        cfg.CONF.set_override('vif_type', 'ovs', 'NOVA')

    def test_default_viftype(self):
        with self.port() as port:
            self.assertEqual(port['port']['binding:vif_type'], 'ovs')


class TestBigSwitchProxyNetworksV2(test_plugin.TestNetworksV2,
                                   BigSwitchProxyPluginV2TestCase):

    def _get_networks(self, tenant_id):
        ctx = context.Context('', tenant_id)
        return directory.get_plugin().get_networks(ctx)

    def test_rollback_on_network_create(self):
        tid = test_base._uuid()
        kwargs = {'tenant_id': tid}
        self.httpPatch.stop()
        with patch(HTTPCON, new=fake_server.HTTPConnectionMock500):
            self._create_network('json', 'netname', True, **kwargs)
        self.httpPatch.start()
        self.assertFalse(self._get_networks(tid))

    def test_rollback_on_network_update(self):
        with self.network() as n:
            data = {'network': {'name': 'aNewName'}}
            self.httpPatch.stop()
            with patch(HTTPCON, new=fake_server.HTTPConnectionMock500):
                self.new_update_request(
                    'networks', data, n['network']['id']
                ).get_response(self.api)
            self.httpPatch.start()
            updatedn = self._get_networks(n['network']['tenant_id'])[0]
            # name should have stayed the same due to failure
            self.assertEqual(n['network']['name'], updatedn['name'])

    def test_rollback_on_network_delete(self):
        with self.network() as n:
            self.httpPatch.stop()
            with patch(HTTPCON, new=fake_server.HTTPConnectionMock500):
                self._delete(
                    'networks', n['network']['id'],
                    expected_code=webob.exc.HTTPInternalServerError.code)
            self.httpPatch.start()
            # network should still exist in db
            self.assertEqual(n['network']['id'],
                             self._get_networks(n['network']['tenant_id']
                                                )[0]['id'])

    def test_notify_on_security_group_change(self):
        plugin = directory.get_plugin()
        with self.port() as p:
            with\
                mock.patch.object(plugin, 'notifier') as n_mock,\
                mock.patch.object(plugin, 'is_security_group_member_updated',
                                  return_value=True):
                # any port update should trigger a notification due to s_mock
                data = {'port': {'name': 'aNewName'}}
                self.new_update_request(
                    'ports', data, p['port']['id']).get_response(self.api)
                self.assertTrue(n_mock.port_update.called)


class TestBigSwitchProxySubnetsV2(test_plugin.TestSubnetsV2,
                                  BigSwitchProxyPluginV2TestCase):

    pass


class TestBigSwitchProxySync(BigSwitchProxyPluginV2TestCase):

    def test_send_data(self):
        plugin_obj = directory.get_plugin()
        result = plugin_obj._send_all_data()
        self.assertEqual(result[0], 200)


class TestDisplayName(BigSwitchProxyPluginV2TestCase):
    def get_true(self):
        """Used for side_effect replacement

        :return:
        """
        return True

    def test_map_display_name_or_tenant_unicode_disabled(self):
        """Test _map_display_name_or_tenant behaviors when unicode is disabled

        :return:
        """
        self.map_display_name_or_tenant_p.stop()
        plugin_obj = directory.get_plugin()

        self.assertFalse(plugin_obj.servers.is_unicode_enabled())

        # object with non-existing tenant_id
        no_tenant_obj = {'id': 'test_id',
                         'name': 'test_name',
                         'tenant_id': 'non_exist_tenant_id'}

        self.assertRaises(TenantIDNotFound,
                          plugin_obj._map_display_name_or_tenant,
                          no_tenant_obj)

        # add a tenant to cache
        plugin_obj.servers.keystone_tenants = {'tenant_id': 'tenant_name'}

        # object with name, '_' in name will be replaced with '__'
        test_obj = {'id': 'test_id',
                    'name': 'test_name',
                    'tenant_id': 'tenant_id'}

        expected_obj = {'id': 'test_id',
                        'name': 'test__name',
                        'tenant_id': 'tenant_id',
                        'tenant_name': 'tenant_name'}

        self.assertEqual(expected_obj,
                         plugin_obj._map_display_name_or_tenant(test_obj))

        # object without name
        test_obj = {'id': 'test_id',
                    'tenant_id': 'tenant_id'}

        expected_obj = {'id': 'test_id',
                        'tenant_id': 'tenant_id',
                        'tenant_name': 'tenant_name'}

        self.assertEqual(expected_obj,
                         plugin_obj._map_display_name_or_tenant(test_obj))

    def test_map_display_name_or_tenant_unicode_enabled(self):
        """Test _map_display_name_or_tenant behaviors when unicode is enabled

        :return:
        """
        self.map_display_name_or_tenant_p.stop()
        self.is_unicode_enabled_p.stop()
        mock.patch(IS_UNICODE_ENABLED, side_effect=self.get_true).start()
        plugin_obj = directory.get_plugin()

        self.assertTrue(plugin_obj.servers.is_unicode_enabled())

        # object with non-existing tenant_id, unicode enabled
        no_tenant_obj = {'id': 'test_id',
                         'name': 'test_name',
                         'tenant_id': 'non_exist_tenant_id'}

        self.assertRaises(TenantIDNotFound,
                          plugin_obj._map_display_name_or_tenant,
                          no_tenant_obj)

        # add a tenant to cache
        plugin_obj.servers.keystone_tenants = {'tenant_id': 'tenant_name'}

        # object with name, unicode enabled
        test_obj = {'id': 'test_id',
                    'name': 'test_name',
                    'tenant_id': 'tenant_id'}

        expected_obj = {'id': 'test_id',
                        'name': 'test_id',
                        'display-name': 'test_name',
                        'tenant_id': 'tenant_id'}

        self.assertEqual(expected_obj,
                         plugin_obj._map_display_name_or_tenant(test_obj))

        # object without name, unicode enabled
        test_obj = {'id': 'test_id',
                    'tenant_id': 'tenant_id'}

        expected_obj = {'id': 'test_id',
                        'name': 'test_id',
                        'tenant_id': 'tenant_id'}

        self.assertEqual(expected_obj,
                         plugin_obj._map_display_name_or_tenant(test_obj))


class TestBigSwitchAddressPairs(test_addr_pair.TestAllowedAddressPairs,
                                BigSwitchProxyPluginV2TestCase):
    def test_create_missing_mac_field(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_create_address_gets_port_mac(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_create_overlap_with_fixed_ip(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_create_port_remove_allowed_address_pairs_with_none(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_create_port_allowed_address_pairs(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_equal_to_max_allowed_address_pair(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_update_port_allowed_address_pairs_bad_format(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_create_port_remove_allowed_address_pairs_with_list(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_update_add_address_pairs(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_update_with_none_and_own_mac_for_duplicate_ip(self):
        # TODO(wolverineav): debug why this fails
        pass

    def test_update_add_address_pairs_with_unexpected_format(self):
        # TODO(wolverineav): debug why this fails
        pass



#!/usr/bin/env python
# Copyright 2012, Big Switch Networks, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test server mocking a REST based network ctrl.

Used for NeutronRestProxy tests
"""
from __future__ import print_function

import re

from oslo_serialization import jsonutils
from six import moves
from wsgiref import simple_server


class TestNetworkCtrl(object):

    def __init__(self, host='', port=8000,
                 default_status='404 Not Found',
                 default_response='404 Not Found',
                 debug=False):
        self.host = host
        self.port = port
        self.default_status = default_status
        self.default_response = default_response
        self.debug = debug
        self.debug_env = False
        self.debug_resp = False
        self.matches = []

    def match(self, prior, method_regexp, uri_regexp, handler, data=None,
              multi=True):
        """Add to the list of expected inputs.

        The incoming request is matched in the order of priority. For same
        priority, match the oldest match request first.

        :param prior: integer priority of this match (e.g. 100)
        :param method_regexp: regexp to match method (e.g. 'PUT|POST')
        :param uri_regexp: regexp to match uri (e.g. '/quantum/v?.?/')
        :param handler: function with signature:
            lambda(method, uri, body, **kwargs) : status, body
              where
                  - method: HTTP method for this request
                  - uri: URI for this HTTP request
                  - body: body of this HTTP request
                  - kwargs are:
                      - data: data object that was in the match call
                      - node: TestNetworkCtrl object itself
                      - id: offset of the matching tuple
              and return values is:
                (status, body) where:
                - status: HTTP resp status (e.g. '200 OK').
                          If None, use default_status
                - body: HTTP resp body. If None, use ''
        """
        assert int(prior) == prior, 'Priority should an integer be >= 0'
        assert prior >= 0, 'Priority should an integer be >= 0'

        lo, hi = 0, len(self.matches)
        while lo < hi:
            mid = (lo + hi) // 2
            if prior < self.matches[mid][0]:
                hi = mid
            else:
                lo = mid + 1
        self.matches.insert(lo, (prior, method_regexp, uri_regexp, handler,
                            data, multi))

    def remove_id(self, id_):
        assert id_ >= 0, 'remove_id: id < 0'
        assert id_ <= len(self.matches), 'remove_id: id > len()'
        self.matches.pop(id_)

    def request_handler(self, method, uri, body):
        retstatus = self.default_status
        retbody = self.default_response
        for i in moves.range(len(self.matches)):
            (unused_prior, method_regexp, uri_regexp, handler, data,
             multi) = self.matches[i]
            if re.match(method_regexp, method) and re.match(uri_regexp, uri):
                kwargs = {
                    'data': data,
                    'node': self,
                    'id': i,
                }
                retstatus, retbody = handler(method, uri, body, **kwargs)
                if multi is False:
                    self.remove_id(i)
                break
        if retbody is None:
            retbody = ''
        return (retstatus, retbody)

    def server(self):
        def app(environ, start_response):
            uri = environ['PATH_INFO']
            method = environ['REQUEST_METHOD']
            headers = [('Content-type', 'text/json')]
            content_len_str = environ['CONTENT_LENGTH']

            content_len = 0
            request_data = None
            if content_len_str:
                content_len = int(content_len_str)
                request_data = environ.get('wsgi.input').read(content_len)
                if request_data:
                    try:
                        request_data = jsonutils.loads(request_data)
                    except Exception:
                        # OK for it not to be json! Ignore it
                        pass

            if self.debug:
                print('\n')
                if self.debug_env:
                    print('environ:')
                    for (key, value) in sorted(environ.items()):
                        print('  %16s : %s' % (key, value))

                print('%s %s' % (method, uri))
                if request_data:
                    print('%s' %
                          jsonutils.dumps(
                              request_data, sort_keys=True, indent=4))

            status, body = self.request_handler(method, uri, None)
            body_data = None
            if body:
                try:
                    body_data = jsonutils.loads(body)
                except Exception:
                    # OK for it not to be json! Ignore it
                    pass

            start_response(status, headers)
            if self.debug:
                if self.debug_env:
                    print('%s: %s' % ('Response',
                          jsonutils.dumps(
                              body_data, sort_keys=True, indent=4)))
            return body
        return simple_server.make_server(self.host, self.port, app)

    def run(self):
        print("Serving on port %d ..." % self.port)
        try:
            self.server().serve_forever()
        except KeyboardInterrupt:
            pass


if __name__ == "__main__":
    import sys

    port = 8899
    if len(sys.argv) > 1:
        port = int(sys.argv[1])

    debug = False
    if len(sys.argv) > 2:
        if sys.argv[2].lower() in ['debug', 'true']:
            debug = True

    ctrl = TestNetworkCtrl(port=port,
                           default_status='200 OK',
                           default_response='{"status":"200 OK"}',
                           debug=debug)
    ctrl.match(100, 'GET', '/test',
               lambda m, u, b, **k: ('200 OK', '["200 OK"]'))
    ctrl.run()



# Copyright 2014 Big Switch Networks, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os

import mock
from neutron.tests.unit.api.v2 import test_base
from neutron.tests.unit.db import test_db_base_plugin_v2 as test_plugin
from oslo_config import cfg
from oslo_log import log as logging

from networking_bigswitch.tests.unit.bigswitch.mock_paths import CERT_COMBINER
from networking_bigswitch.tests.unit.bigswitch.mock_paths import FILE_PUT
from networking_bigswitch.tests.unit.bigswitch.mock_paths import GET_CA_CERTS
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    GET_SERVER_CERTIFICATE
from networking_bigswitch.tests.unit.bigswitch.mock_paths import HTTPSCON
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    SSL_CREATE_DEFAULT_CONTEXT
from networking_bigswitch.tests.unit.bigswitch \
    import test_base as bsn_test_base

LOG = logging.getLogger(__name__)


class test_ssl_certificate_base(test_plugin.NeutronDbPluginV2TestCase,
                                bsn_test_base.BigSwitchTestBase):

    plugin_str = ('%s.NeutronRestProxyV2' %
                  bsn_test_base.PLUGIN_PATH)
    servername = None
    cert_base = None

    def _setUp(self):
        self.servername = test_base._uuid()
        self.cert_base = cfg.CONF.RESTPROXY.ssl_cert_directory
        self.host_cert_val = 'DUMMYCERTFORHOST%s' % self.servername
        self.host_cert_path = os.path.join(
            self.cert_base,
            'host_certs',
            '%s.pem' % self.servername
        )
        self.comb_cert_path = os.path.join(
            self.cert_base,
            'combined',
            '%s.pem' % self.servername
        )
        self.ca_certs_path = os.path.join(
            self.cert_base,
            'ca_certs'
        )
        cfg.CONF.set_override('servers', ["%s:443" % self.servername],
                              'RESTPROXY')
        self.setup_patches()

        # Mock method SSL lib uses to grab cert from server
        self.sslgetcert_m = mock.patch(GET_SERVER_CERTIFICATE,
                                       create=True).start()
        self.sslgetcert_m.return_value = self.host_cert_val

        # Mock methods that write and read certs from the file-system
        self.fileput_m = mock.patch(FILE_PUT, create=True).start()
        self.certcomb_m = mock.patch(CERT_COMBINER, create=True).start()
        self.getcacerts_m = mock.patch(GET_CA_CERTS, create=True).start()

    def setUp(self):
        test_plugin.NeutronDbPluginV2TestCase.setUp(self, self.plugin_str)
        self.setup_db()


class TestSslSticky(test_ssl_certificate_base):

    def setUp(self):
        self.setup_config_files()
        cfg.CONF.set_override('server_ssl', True, 'RESTPROXY')
        cfg.CONF.set_override('ssl_sticky', True, 'RESTPROXY')
        self._setUp()
        # # Set fake HTTPS connection's expectation
        # self.fake_certget_m.return_value = self.host_cert_val
        # No CA certs for this test
        self.getcacerts_m.return_value = []

        self.ssl_create_default_context_m = mock.patch(
            SSL_CREATE_DEFAULT_CONTEXT,
            create=True).start()
        super(TestSslSticky, self).setUp()

    def test_sticky_cert(self):
        """Test certificate is stored, and https connection is used

        :return:
        """
        # SSL connection should be successful and cert should be cached
        with mock.patch(HTTPSCON) as https_mock, self.network():
            # CA certs should have been checked for
            self.getcacerts_m.assert_has_calls([mock.call(self.ca_certs_path)])
            # cert should have been fetched via SSL lib
            self.sslgetcert_m.assert_has_calls(
                [mock.call((self.servername, 443))]
            )

            # cert should have been recorded
            self.fileput_m.assert_has_calls([mock.call(self.host_cert_path,
                                                       self.host_cert_val)])

            # no ca certs, so host cert only for this combined cert
            self.certcomb_m.assert_has_calls([mock.call([self.host_cert_path],
                                                        self.comb_cert_path)])

            # confirm that ssl_context is created
            self.ssl_create_default_context_m.assert_called_once()

            # Test HTTPS Connection is used for REST Calls
            https_mock.assert_called_once()

            # confirm that ssl_context is passed
            self.assertTrue(https_mock.
                            call_args_list[0][1].get('context'))



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""update column type for testpath result

Revision ID: 1086543dfc0f
Revises: kilo
Create Date: 2016-04-25 11:54:44.646798

"""

from alembic import op
from sqlalchemy.types import Enum

# revision identifiers, used by Alembic.
revision = '1086543dfc0f'
down_revision = 'kilo'
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column(
        'reachabilitytest', 'expected_result',
        type_=Enum("dropped by route",
                   "dropped by policy",
                   "not permitted by security groups",
                   "dropped due to private segment",
                   "dropped due to loop",
                   "packet in",
                   "forwarded",
                   "dropped",
                   "unspecified source",
                   "unsupported",
                   "invalid input",
                   "inconsistent status",
                   "no traffic detected",
                   name="expected_result"),
        existing_nullable=False)

    op.alter_column(
        'reachabilityquicktest', 'expected_result',
        type_=Enum("dropped by route",
                   "dropped by policy",
                   "not permitted by security groups",
                   "dropped due to private segment",
                   "dropped due to loop",
                   "packet in",
                   "forwarded",
                   "dropped",
                   "unspecified source",
                   "unsupported",
                   "invalid input",
                   "inconsistent status",
                   "no traffic detected",
                   name="expected_result"),
        existing_nullable=False)


def downgrade():
    pass



# Copyright 2017 Big Switch Networks, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os

from networking_bigswitch.plugins.bigswitch import constants as bsn_consts
from networking_bigswitch.plugins.bigswitch.i18n import _LI
from networking_bigswitch.plugins.bigswitch.i18n import _LW
from neutron_lib import exceptions as n_exc
from oslo_log import log

LOG = log.getLogger(__name__)

from networking_bigswitch.plugins.bigswitch.i18n import _


class Util(object):
    """Util

    Placeholder for static methods that can be called from across the plugin
    and reused as required.
    """

    @staticmethod
    def format_resource_name(name):
        """format resource name

        Util method to format resource names to make them compatible with BCF.

        Replaces special characters with its corresponding BCF compatible
        encoding.

        :param name non empty string which is the name of the resource
        :rtype string name with special characters replaced
        """
        return (name
                # always replace underscores first, since other replacements
                # contain underscores as part of replacement
                .replace('_', '__')
                .replace(' ', '_s')
                .replace('\'', '_a')
                .replace('/', '_f')
                .replace('[', '_l')
                .replace(']', '_r'))

    @staticmethod
    def get_tenant_id_for_create(context, resource):
        if context.is_admin and 'tenant_id' in resource:
            tenant_id = resource['tenant_id']
        elif ('tenant_id' in resource and
              resource['tenant_id'] != context.tenant_id):
            reason = _('Cannot create resource for another tenant')
            raise n_exc.AdminRequired(reason=reason)
        else:
            tenant_id = context.tenant_id
        return tenant_id

    @staticmethod
    def read_ovs_bridge_mappings():
        """Read the 'bridge_mappings' property from openvswitch_agent.ini

        This is done for Redhat environments, to allow an improved
        learning/programming of interface groups based on ports and the network
        to which the ports belong to.

        :return: bridge_mappings dictionary {'physnet_name': 'bridge_name', ..}
                    {} empty dictionary when not found
        """
        mapping = {}
        mapping_str = None
        # read openvswitch_agent.ini for bridge_mapping info
        if not os.path.isfile(bsn_consts.OVS_AGENT_INI_FILEPATH):
            # if ovs_agent.ini doesn't exists, return empty mapping
            LOG.warning(_LW("Unable to read OVS bridge_mappings, "
                            "openvswitch_agent.ini file not present."))
            return mapping

        with open(bsn_consts.OVS_AGENT_INI_FILEPATH) as f:
            for line in f:
                if ('#' not in line and
                        ('=' in line and 'bridge_mappings' in line)):
                    # typical config line looks like the following:
                    # bridge_mappings = datacentre:br-ex,dpdk:br-link
                    key, value = line.split('=', 1)
                    mapping_str = value.strip()

        # parse comma separated physnet list into individual mappings
        if not mapping_str:
            # if file did not have bridge_mappings, return empty mapping
            LOG.warning(_LW(
                "Unable to read OVS bridge_mappings, either the line is "
                "commented or not present in openvswitch_agent.ini."))
            return mapping

        phy_map_list = mapping_str.split(',')
        for phy_map in phy_map_list:
            phy, bridge = phy_map.split(':')
            mapping[phy.strip()] = bridge.strip()

        LOG.info(_LI("OVS bridge_mappings are: %(br_map)s"),
                 {'br_map': mapping})
        return mapping



# Copyright 2014 Big Switch Networks, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket
import time

import mock
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_utils import importutils
from six.moves import http_client

from networking_bigswitch.plugins.bigswitch.db import consistency_db
from networking_bigswitch.plugins.bigswitch import servermanager
from networking_bigswitch.tests.unit.bigswitch.mock_paths import CERT_COMBINER
from networking_bigswitch.tests.unit.bigswitch.mock_paths import CONSISTENCY_DB
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    GET_SERVER_CERTIFICATE
from networking_bigswitch.tests.unit.bigswitch.mock_paths import HTTPCON
from networking_bigswitch.tests.unit.bigswitch.mock_paths import HTTPSCON
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    POOL_GET_CAPABILITIES
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    POOL_REST_ACTION
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    POOL_REST_CALL
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    POOL_TOPO_SYNC
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    POOL_UPDATE_TENANT_CACHE
from networking_bigswitch.tests.unit.bigswitch.mock_paths import SERVER_MANAGER
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    SERVER_REST_CALL
from networking_bigswitch.tests.unit.bigswitch \
    import test_restproxy_plugin as test_rp


class ServerManagerTests(test_rp.BigSwitchProxyPluginV2TestCase):

    def setUp(self):
        self.socket_mock = mock.patch(
            SERVER_MANAGER + '.socket.create_connection').start()
        self.wrap_mock = mock.patch(SERVER_MANAGER +
                                    '.ssl.wrap_socket').start()
        super(ServerManagerTests, self).setUp()
        # http patch must not be running or it will mangle the servermanager
        # import where the https connection classes are defined
        self.httpPatch.stop()
        self.sm = importutils.import_module(SERVER_MANAGER)

    def test_no_servers(self):
        cfg.CONF.set_override('servers', [], 'RESTPROXY')
        self.assertRaises(cfg.Error, servermanager.ServerPool)

    def test_malformed_servers(self):
        cfg.CONF.set_override('servers', ['1.2.3.4', '1.1.1.1:a'], 'RESTPROXY')
        self.assertRaises(cfg.Error, servermanager.ServerPool)

    def test_ipv6_server_address(self):
        cfg.CONF.set_override(
            'servers', ['[ABCD:EF01:2345:6789:ABCD:EF01:2345:6789]:80'],
            'RESTPROXY')
        s = servermanager.ServerPool()
        self.assertEqual(s.servers[0].server,
                         'ABCD:EF01:2345:6789:ABCD:EF01:2345:6789')

    def test_sticky_cert_fetch_fail(self):
        pl = directory.get_plugin()
        pl.servers.is_ssl_enabled = True

        with mock.patch(GET_SERVER_CERTIFICATE,
                        side_effect=Exception('There is no more entropy in the'
                                              'universe')) as sslgetmock:
            self.assertRaises(
                cfg.Error,
                pl.servers._get_combined_cert_for_server,
                *('example.org', 443)
            )
            sslgetmock.assert_has_calls([mock.call(('example.org', 443))])

    def test_consistency_watchdog_stops_with_0_polling_interval(self):
        pl = directory.get_plugin()
        pl.servers.capabilities = ['consistency']
        self.watch_p.stop()
        with mock.patch('eventlet.sleep') as smock:
            # should return immediately a polling interval of 0
            pl.servers._consistency_watchdog(0)
            self.assertFalse(smock.called)

    def test_consistency_watchdog(self):
        pl = directory.get_plugin()
        pl.servers.capabilities = ['dummy']
        self.watch_p.stop()

        with mock.patch('eventlet.sleep') as smock,\
                mock.patch(
                    POOL_REST_CALL,
                    side_effect=servermanager.RemoteRestError(
                        reason='Failure to trigger except clause.'))\
                as rmock,\
                mock.patch(
                    SERVER_MANAGER + '.LOG.exception',
                    side_effect=KeyError('Failure to break loop'))\
                as lmock:
            # should return immediately without consistency capability
            pl.servers._consistency_watchdog()
            self.assertFalse(smock.called)

            pl.servers.capabilities = ['consistency']
            self.assertRaises(KeyError,
                              pl.servers._consistency_watchdog)
            rmock.assert_called_with('GET', '/health', '', {}, [], False)
            self.assertEqual(1, len(lmock.mock_calls))

    def test_file_put_contents(self):
        pl = directory.get_plugin()
        with mock.patch(SERVER_MANAGER + '.open', create=True) as omock:
            pl.servers._file_put_contents('somepath', 'contents')
            omock.assert_has_calls([mock.call('somepath', 'w')])
            omock.return_value.__enter__.return_value.assert_has_calls([
                mock.call.write('contents')
            ])

    def test_combine_certs_to_file(self):
        pl = directory.get_plugin()
        with mock.patch(SERVER_MANAGER + '.open', create=True) as omock:
            omock.return_value.__enter__().read.return_value = 'certdata'
            pl.servers._combine_certs_to_file(['cert1.pem', 'cert2.pem'],
                                              'combined.pem')
            # mock shared between read and write file handles so the calls
            # are mixed together
            omock.assert_has_calls([
                mock.call('combined.pem', 'w'),
                mock.call('cert1.pem', 'r'),
                mock.call('cert2.pem', 'r'),
            ], any_order=True)
            omock.return_value.__enter__.return_value.assert_has_calls([
                mock.call.read(),
                mock.call.write('certdata'),
                mock.call.read(),
                mock.call.write('certdata')
            ])

    # basic authentication
    def test_auth_header(self):
        cfg.CONF.set_override('server_auth', 'username:pass', 'RESTPROXY')
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value.getheader.return_value = 'HASHHEADER'
            sp.rest_create_network('tenant', 'network')
        callheaders = rv.request.mock_calls[0][1][3]
        self.assertIn('Authorization', callheaders)
        self.assertNotIn('Cookie', callheaders)
        self.assertEqual(callheaders['Authorization'],
                         'Basic dXNlcm5hbWU6cGFzcw==')

    # token based authentication
    def test_auth_token_header(self):
        cfg.CONF.set_override('server_auth', 'fake_token', 'RESTPROXY')
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value.getheader.return_value = 'HASHHEADER'
            sp.rest_create_network('tenant', 'network')
        callheaders = rv.request.mock_calls[0][1][3]
        self.assertIn('Cookie', callheaders)
        self.assertNotIn('Authorization', callheaders)
        self.assertEqual(callheaders['Cookie'], 'session_cookie="fake_token"')

    def test_header_add(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value.getheader.return_value = 'HASHHEADER'
            sp.servers[0].rest_call('GET', '/', headers={'EXTRA-HEADER': 'HI'})
        callheaders = rv.request.mock_calls[0][1][3]
        # verify normal headers weren't mangled
        self.assertIn('Content-type', callheaders)
        self.assertEqual(callheaders['Content-type'],
                         'application/json')
        # verify new header made it in
        self.assertIn('EXTRA-HEADER', callheaders)
        self.assertEqual(callheaders['EXTRA-HEADER'], 'HI')

    def test_capabilities_retrieval(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value.getresponse.return_value
            rv.getheader.return_value = 'HASHHEADER'

            # each server will get different capabilities
            rv.read.side_effect = ['["a","b","c"]', '["b","c","d"]']
            # pool capabilities is union of both
            # normally capabilities should be the same across all servers
            # this only happens in two situations:
            # 1. a server is down
            # 2. during upgrade/downgrade
            self.assertEqual(set(['a', 'b', 'c', 'd']), sp.get_capabilities())
            self.assertEqual(2, rv.read.call_count)

            # the pool should cache after the first call
            # so no more HTTP calls should be made
            rv.read.side_effect = ['["w","x","y"]', '["x","y","z"]']
            self.assertEqual(set(['a', 'b', 'c', 'd']), sp.get_capabilities())
            self.assertEqual(2, rv.read.call_count)

    def test_capabilities_retrieval_failure(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value.getresponse.return_value
            rv.getheader.return_value = 'HASHHEADER'
            # a failure to parse should result in an empty capability set
            rv.read.return_value = 'XXXXX'
            self.assertEqual([], sp.servers[0].get_capabilities())

            # as capabilities is empty, it should try to update capabilities
            rv.read.side_effect = ['{"a": "b"}', '["b","c","d"]']
            self.assertEqual(set(['a', 'b', 'c', 'd']), sp.get_capabilities())

    def test_reconnect_on_timeout_change(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value.getheader.return_value = 'HASHHEADER'
            sp.servers[0].capabilities = ['keep-alive']
            sp.servers[0].rest_call('GET', '/', timeout=10)
            # even with keep-alive enabled, a change in timeout will trigger
            # a reconnect
            sp.servers[0].rest_call('GET', '/', timeout=75)
        conmock.assert_has_calls([
            mock.call('localhost', 9000, timeout=10),
            mock.call('localhost', 9000, timeout=75),
        ], any_order=True)

    def test_connect_failures(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON, return_value=None):
            resp = sp.servers[0].rest_call('GET', '/')
            self.assertEqual(resp, (0, None, None, None))
        # verify same behavior on ssl class
        sp.servers[0].currentcon = False
        sp.servers[0].is_ssl_enabled = True
        with mock.patch(HTTPSCON, return_value=None):
            resp = sp.servers[0].rest_call('GET', '/')
            self.assertEqual(resp, (0, None, None, None))

    def test_reconnect_cached_connection(self):
        self.skipTest("cached connections are currently disabled because "
                      "their assignment to the servermanager object is not "
                      "thread-safe")
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value.getheader.return_value = 'HASH'
            sp.servers[0].capabilities = ['keep-alive']
            sp.servers[0].rest_call('GET', '/first')
            # raise an error on re-use to verify reconnect
            # return okay the second time so the reconnect works
            rv.request.side_effect = [http_client.ImproperConnectionState(),
                                      mock.MagicMock()]
            sp.servers[0].rest_call('GET', '/second')
        uris = [c[1][1] for c in rv.request.mock_calls]
        expected = [
            sp.base_uri + '/first',
            sp.base_uri + '/second',
            sp.base_uri + '/second',
        ]
        self.assertEqual(uris, expected)

    def test_no_reconnect_recurse_to_infinity(self):
        self.skipTest("cached connections are currently disabled because "
                      "their assignment to the servermanager object is not "
                      "thread-safe")
        # retry uses recursion when a reconnect is necessary
        # this test makes sure it stops after 1 recursive call
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            # hash header must be string instead of mock object
            rv.getresponse.return_value.getheader.return_value = 'HASH'
            sp.servers[0].capabilities = ['keep-alive']
            sp.servers[0].rest_call('GET', '/first')
            # after retrying once, the rest call should raise the
            # exception up
            rv.request.side_effect = http_client.ImproperConnectionState()
            self.assertRaises(http_client.ImproperConnectionState,
                              sp.servers[0].rest_call,
                              *('GET', '/second'))
            # 1 for the first call, 2 for the second with retry
            self.assertEqual(rv.request.call_count, 3)

    def test_socket_error(self):
        sp = servermanager.ServerPool()
        with mock.patch(HTTPCON) as conmock:
            conmock.return_value.request.side_effect = socket.timeout()
            resp = sp.servers[0].rest_call('GET', '/')
            self.assertEqual(resp, (0, None, None, None))

    def test_cert_get_fail(self):
        pl = directory.get_plugin()
        pl.servers.is_ssl_enabled = True
        with mock.patch('os.path.exists', return_value=False):
            self.assertRaises(cfg.Error,
                              pl.servers._get_combined_cert_for_server,
                              *('example.org', 443))

    def test_cert_make_dirs(self):
        pl = directory.get_plugin()
        pl.servers.is_ssl_enabled = True
        cfg.CONF.set_override('ssl_sticky', False, 'RESTPROXY')
        # pretend base dir exists, 3 children don't, and host cert does
        with mock.patch('os.path.exists',
                        side_effect=[True, False, False,
                                     False, True]) as exmock,\
                mock.patch('os.makedirs') as makemock,\
                mock.patch(CERT_COMBINER) as combmock:
            # will raise error because no certs found
            self.assertIn(
                'example.org',
                pl.servers._get_combined_cert_for_server('example.org', 443)
            )
            base = cfg.CONF.RESTPROXY.ssl_cert_directory
            hpath = base + '/host_certs/example.org.pem'
            combpath = base + '/combined/example.org.pem'
            combmock.assert_has_calls([mock.call([hpath], combpath)])
            self.assertEqual(exmock.call_count, 5)
            self.assertEqual(makemock.call_count, 3)

    def test_no_cert_error(self):
        pl = directory.get_plugin()
        pl.servers.is_ssl_enabled = True
        cfg.CONF.set_override('ssl_sticky', False, 'RESTPROXY')
        # pretend base dir exists and 3 children do, but host cert doesn't
        with mock.patch(
            'os.path.exists',
            side_effect=[True, True, True, True, False]
        ) as exmock:
            # will raise error because no certs found
            self.assertRaises(
                cfg.Error,
                pl.servers._get_combined_cert_for_server,
                *('example.org', 443)
            )
            self.assertEqual(exmock.call_count, 5)

    def test_action_success(self):
        pl = directory.get_plugin()
        self.assertTrue(pl.servers.action_success((200,)))

    def test_server_failure(self):
        pl = directory.get_plugin()
        self.assertTrue(pl.servers.server_failure((404,)))
        # server failure has an ignore codes option
        self.assertFalse(pl.servers.server_failure((404,),
                                                   ignore_codes=[404]))

    def test_retry_on_unavailable(self):
        pl = directory.get_plugin()
        with mock.patch(SERVER_REST_CALL,
                        return_value=(http_client.SERVICE_UNAVAILABLE,
                                      0, 0, 0)) as srestmock,\
                mock.patch(SERVER_MANAGER + '.eventlet.sleep') as tmock:
            # making a call should trigger retries with sleeps in between
            pl.servers.rest_call('GET', '/', '', None, [])
            rest_call = [mock.call('GET', '/', '', None, False,
                                   reconnect=True)]
            rest_call_count = (
                servermanager.HTTP_SERVICE_UNAVAILABLE_RETRY_COUNT + 1)
            srestmock.assert_has_calls(rest_call * rest_call_count)
            sleep_call = [mock.call(
                servermanager.HTTP_SERVICE_UNAVAILABLE_RETRY_INTERVAL)]
            # should sleep 1 less time than the number of calls
            sleep_call_count = rest_call_count - 1
            tmock.assert_has_calls(sleep_call * sleep_call_count)

    def test_delete_failure_forces_topo_sync(self):
        pl = directory.get_plugin()
        with mock.patch(SERVER_REST_CALL,
                        return_value=(http_client.INTERNAL_SERVER_ERROR,
                                      0, 0, 0)), \
                mock.patch(POOL_TOPO_SYNC,
                           return_value=(False,
                                         servermanager.TOPO_RESPONSE_OK)) \
                as topo_mock:
            # a failed DELETE call should trigger a forced topo_sync
            # with check_ts True
            self.assertRaises(servermanager.RemoteRestError,
                              pl.servers.rest_action,
                              **{'action': 'DELETE', 'resource': '/',
                                 'data': '',
                                 'errstr': "Unable to DELETE query to CCF: %s",
                                 'ignore_codes': []})
            topo_mock.assert_called_once_with(**{'check_ts': True})

    def test_post_failure_forces_topo_sync(self):
        pl = directory.get_plugin()
        with mock.patch(SERVER_REST_CALL,
                        return_value=(http_client.INTERNAL_SERVER_ERROR,
                                      0, 0, 0)), \
                mock.patch(POOL_TOPO_SYNC,
                           return_value=(False,
                                         servermanager.TOPO_RESPONSE_OK)) \
                as topo_mock:
            # a failed POST call should trigger a forced topo_sync
            # with check_ts True
            self.assertRaises(servermanager.RemoteRestError,
                              pl.servers.rest_action,
                              **{'action': 'POST', 'resource': '/', 'data': '',
                                 'errstr': "Unable to POST query to CCF: %s",
                                 'ignore_codes': []})
            topo_mock.assert_called_once_with(**{'check_ts': True})

    def test_topo_sync_failure_does_not_force_topo_sync(self):
        pl = directory.get_plugin()
        with mock.patch(SERVER_REST_CALL,
                        return_value=(http_client.INTERNAL_SERVER_ERROR,
                                      0, 0, 0)), \
                mock.patch(POOL_TOPO_SYNC,
                           return_value=(False,
                                         servermanager.TOPO_RESPONSE_OK)) \
                as topo_mock:
            # a failed POST call for topology path should raise an exception
            # and not call force_topo_sync like other failed rest_action
            self.assertRaises(servermanager.RemoteRestError,
                              pl.servers.rest_action,
                              **{'action': 'POST', 'resource': '/topology',
                                 'data': '',
                                 'errstr': "Unable to perform topo_sync: %s",
                                 'ignore_codes': []})
            topo_mock.assert_not_called()

    def test_not_found_sync_raises_error_without_topology(self):
        pl = directory.get_plugin()
        pl.servers.get_topo_function = None
        with \
            mock.patch(
                SERVER_REST_CALL,
                return_value=(http_client.NOT_FOUND, 0, 0, 0)):
            # making a call should trigger a conflict sync that will
            # error without the topology function set
            self.assertRaises(
                cfg.Error,
                pl.servers.rest_action,
                *('GET', '/', '', None, [])
            )

    def test_no_sync_without_keystone(self):
        pl = directory.get_plugin()
        with\
            mock.patch(POOL_UPDATE_TENANT_CACHE, return_value=(False)),\
            mock.patch(SERVER_REST_CALL,
                       return_value=(http_client.CONFLICT, 0, 0,
                                     0)) as srestmock:
            # making a call should trigger a conflict sync
            pl.servers.rest_call('GET', '/', '', None, [])
            srestmock.assert_called_once_with(
                'GET', '/', '', None, False, reconnect=True)

    def test_no_send_all_data_without_keystone(self):
        pl = directory.get_plugin()
        with mock.patch(POOL_UPDATE_TENANT_CACHE, return_value=(False)), \
            mock.patch(POOL_TOPO_SYNC,
                       return_value=(False, servermanager.TOPO_RESPONSE_OK)) \
                as tmock:
            # making a call should trigger a conflict sync
            self.assertRaises(Exception, pl._send_all_data())  # noqa
            tmock.assert_called_once()

    def test_floating_calls(self):
        pl = directory.get_plugin()
        with mock.patch(POOL_REST_ACTION) as ramock:
            body1 = {'id': 'somefloat'}
            body2 = {'name': 'myfl'}
            pl.servers.rest_create_floatingip('tenant', body1)
            pl.servers.rest_update_floatingip('tenant', body2, 'id')
            pl.servers.rest_delete_floatingip('tenant', 'oldid')
            ramock.assert_has_calls([
                mock.call('PUT', '/tenants/tenant/floatingips/somefloat',
                          body1,
                          errstr=u'Unable to create floating IP: %s'),
                mock.call('PUT', '/tenants/tenant/floatingips/id',
                          body2,
                          errstr=u'Unable to update floating IP: %s'),
                mock.call('DELETE', '/tenants/tenant/floatingips/oldid',
                          errstr=u'Unable to delete floating IP: %s')
            ])

    def test_is_unicode_enabled(self):
        """Verify that unicode is enabled only when both conditions are True:

         1. naming_scheme_unicode is True or empty
         2. BCF capabilities include display-name

        :return:
        """
        self.is_unicode_enabled_p.stop()

        def capability_unicode_supported():
            return ['dummy', 'display-name']

        def capability_unicode_unsupported():
            return ['dummy']

        patch_supported = mock.patch(
            POOL_GET_CAPABILITIES,
            side_effect=capability_unicode_supported)

        patch_unsupported = mock.patch(
            POOL_GET_CAPABILITIES,
            side_effect=capability_unicode_unsupported)

        # Create a server pool with default naming_scheme_unicode
        # verify default value is true
        sp = servermanager.ServerPool()
        self.assertTrue(cfg.CONF.RESTPROXY.naming_scheme_unicode)

        # config enabled, and unicode is supported on bcf
        patch_supported.start()
        self.assertTrue(sp.is_unicode_enabled())
        patch_supported.stop()

        # config enabled, but unicode is not supported on bcf
        patch_unsupported.start()
        self.assertFalse(sp.is_unicode_enabled())
        patch_unsupported.stop()

        # Recreate the server pool, as the config is read during initialization
        cfg.CONF.set_override('naming_scheme_unicode', False, 'RESTPROXY')
        sp = servermanager.ServerPool()

        # config disabled, though unicode is supported on bcf
        patch_supported.start()
        self.assertFalse(sp.is_unicode_enabled())
        patch_supported.stop()

        # config disabled, and unicode is not supported on bcf
        patch_unsupported.start()
        self.assertFalse(sp.is_unicode_enabled())
        patch_unsupported.stop()


class HashLockingTests(test_rp.BigSwitchProxyPluginV2TestCase):

    def _get_hash_from_handler_db(self, handler):
        with handler.session.begin(subtransactions=True):
            res = (handler.session.query(consistency_db.ConsistencyHash).
                   filter_by(hash_id=handler.hash_id).first())
            return res.hash

    def test_lock_no_initial_record(self):
        handler = consistency_db.HashHandler()
        h1 = handler.lock()
        # lock() request on empty DB should succeed
        self.assertTrue(h1)
        # db should have a lock marker
        self.assertEqual(handler.lock_marker,
                         self._get_hash_from_handler_db(handler))
        # prev_lock_ts must be 0 for initial case
        self.assertEqual(handler.prev_lock_ts, '0')
        # unlock() should clear the lock
        handler.unlock()
        self.assertEqual(handler.lock_ts,
                         self._get_hash_from_handler_db(handler))

    def test_db_duplicate_on_insert(self):
        handler = consistency_db.HashHandler()
        with mock.patch.object(
            handler.session, 'add', side_effect=[db_exc.DBDuplicateEntry, '']
        ) as add_mock:
            handler.lock()
            # duplicate insert failure should result in retry
            self.assertEqual(2, add_mock.call_count)

    def test_lock_check_ts_true_prev_lock_exists(self):
        handler1 = consistency_db.HashHandler()
        h1 = handler1.lock()
        self.assertTrue(h1)
        self.assertEqual(handler1.lock_marker,
                         self._get_hash_from_handler_db(handler1))

        # 2nd thread came in just 10 millisecs after first one, and first one
        # still holds the lock, expired  = False
        timestamp_2 = float(handler1.lock_ts) + 10
        handler2 = consistency_db.HashHandler(timestamp_ms=timestamp_2)
        h2 = handler2.lock()
        self.assertFalse(h2)
        self.assertEqual(handler1.lock_ts, handler2.prev_lock_ts)

    def test_lock_check_ts_false_prev_lock_exists(self):
        handler1 = consistency_db.HashHandler()
        h1 = handler1.lock()
        self.assertTrue(h1)
        self.assertEqual(handler1.lock_marker,
                         self._get_hash_from_handler_db(handler1))
        self.assertEqual('0', handler1.prev_lock_ts)

        hh2_ts_hh1_ts_plus_1780 = float(handler1.lock_ts) + 1780
        handler2 = consistency_db.HashHandler(
            hash_id='1', timestamp_ms=hh2_ts_hh1_ts_plus_1780)
        with mock.patch(CONSISTENCY_DB + '.eventlet.sleep',
                        side_effect=[Exception]) as emock:
            try:
                handler2.lock(check_ts=False)
            except Exception:
                pass
        self.assertEqual(1, emock.call_count)
        self.assertEqual(handler1.lock_ts, handler2.prev_lock_ts)

    def test_lock_check_ts_true_prev_lock_not_expired(self):
        handler1 = consistency_db.HashHandler()
        h1 = handler1.lock()
        self.assertTrue(h1)
        self.assertEqual(handler1.lock_marker,
                         self._get_hash_from_handler_db(handler1))

        handler1.unlock()
        self.assertEqual(handler1.lock_ts,
                         self._get_hash_from_handler_db(handler1))

        # thread 1 has executed the complete lock-unlock cycle
        # thread 2 now tries to get lock with check_ts True
        # TOPO_SYNC_EXPIRED_SECS = 1800
        hh2_ts_under_limit = float(handler1.lock_ts) + 1000
        handler2 = consistency_db.HashHandler(hash_id=1,
                                              timestamp_ms=hh2_ts_under_limit)
        h2 = handler2.lock()
        self.assertFalse(h2)
        self.assertEqual(handler1.lock_ts, handler2.prev_lock_ts)

    def test_lock_check_ts_true_prev_lock_expired(self):
        handler1 = consistency_db.HashHandler()
        h1 = handler1.lock()
        self.assertTrue(h1)
        self.assertEqual(handler1.lock_marker,
                         self._get_hash_from_handler_db(handler1))

        handler1.unlock()
        self.assertEqual(handler1.lock_ts,
                         self._get_hash_from_handler_db(handler1))

        # thread 1 has executed the complete lock-unlock cycle
        # thread 2 now tries to get lock with check_ts True
        # TOPO_SYNC_EXPIRED_SECS = 1 for testing
        time.sleep(1)
        handler2 = consistency_db.HashHandler()
        # only for testing
        consistency_db.TOPO_SYNC_EXPIRED_SECS = 1
        h2 = handler2.lock()
        self.assertTrue(h2)
        self.assertEqual(handler1.lock_ts, handler2.prev_lock_ts)

    def test_lock_check_ts_false_prev_lock_not_expired(self):
        handler1 = consistency_db.HashHandler()
        h1 = handler1.lock()
        self.assertTrue(h1)
        self.assertEqual(handler1.lock_marker,
                         self._get_hash_from_handler_db(handler1))

        handler1.unlock()
        self.assertEqual(handler1.lock_ts,
                         self._get_hash_from_handler_db(handler1))

        # thread 1 has executed the complete lock-unlock cycle
        # thread 2 now tries to get lock with check_ts True
        # TOPO_SYNC_EXPIRED_SECS = 1800
        hh2_ts_under_limit = float(handler1.lock_ts) + 1000
        handler2 = consistency_db.HashHandler(hash_id=1,
                                              timestamp_ms=hh2_ts_under_limit)
        h2 = handler2.lock(check_ts=False)
        self.assertTrue(h2)
        self.assertEqual(handler1.lock_ts, handler2.prev_lock_ts)

    def test_lock_check_ts_false_lock_clash(self):
        # 2 threads try to lock the DB at the same time when check_ts is False
        # and no thread holds the lock
        handler1 = consistency_db.HashHandler()
        h1 = handler1.lock()
        self.assertTrue(h1)
        handler1.unlock()
        self.assertEqual(handler1.lock_ts,
                         self._get_hash_from_handler_db(handler1))

        handler2 = consistency_db.HashHandler()

        with mock.patch.object(handler2._FACADE, 'get_engine') as ge, \
                mock.patch(CONSISTENCY_DB + '.eventlet.sleep',
                           side_effect=[None]) as emock:
            conn = ge.return_value.begin.return_value.__enter__.return_value
            firstresult = mock.Mock()
            # a rowcount of 0 simulates the effect of another db client
            # updating the same record the handler was trying to update
            firstresult.rowcount = 0
            secondresult = mock.Mock()
            secondresult.rowcount = 1
            conn.execute.side_effect = [firstresult, secondresult]
            h2 = handler2.lock(check_ts=False)
            self.assertTrue(h2)
            # update should have been called again after the failure
            self.assertEqual(2, conn.execute.call_count)
            # sleep should have been called once, during first result failure
            emock.assert_called_once()

    def test_clear_lock(self):
        handler = consistency_db.HashHandler()
        handler.lock()  # lock the table
        self.assertEqual(handler.lock_marker,
                         self._get_hash_from_handler_db(handler))
        handler.unlock()
        self.assertEqual(handler.lock_ts,
                         self._get_hash_from_handler_db(handler))

    def test_handler_already_holding_lock(self):
        handler = consistency_db.HashHandler()
        handler.lock()  # lock the table
        with mock.patch.object(handler._FACADE, 'get_engine') as ge:
            handler.lock()
            # get engine should not have been called because no update
            # should have been made
            self.assertFalse(ge.called)
            self.assertTrue(handler.lock_ts, handler.prev_lock_ts)

    def test_unlock_set_prev_ts(self):
        handler1 = consistency_db.HashHandler()
        handler1.lock()
        self.assertEqual(handler1.lock_marker,
                         self._get_hash_from_handler_db(handler1))
        handler1.unlock()

        # first lock-unlock is done. now comes a second call with
        # check_ts = False
        handler2 = consistency_db.HashHandler()
        h2 = handler2.lock(check_ts=False)
        self.assertTrue(h2)
        self.assertEqual(handler1.lock_ts, handler2.prev_lock_ts)

        # now assuming exception occured during topo_sync, call
        # handler2.unlock(set_prev_ts=True)
        handler2.unlock(set_prev_ts=True)
        # hash in consistency_db will be previous hash_handler's lock_ts
        self.assertEqual(handler1.lock_ts,
                         self._get_hash_from_handler_db(handler2))
        # try unlock again on the same handler2 - it should have no effect
        # as unlock(set_prev_ts) removed TOPO_SYNC marker. this simulates
        # unlock() being called in the finally block of force_topo_sync()
        handler2.unlock()
        self.assertEqual(handler1.lock_ts,
                         self._get_hash_from_handler_db(handler2))



# Copyright 2017, Big Switch Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

BSN_SERVICE_PLUGIN = 'BSNSERVICEPLUGIN'

# state_syncer constants
NOVA_CLIENT_VERSION = '2'
INTF_MAC_ADDR = 'OS-EXT-IPS-MAC:mac_addr'
INTF_IP_ADDR = 'addr'
INTF_IP_TYPE = 'OS-EXT-IPS:type'
INTF_IP_VERSION = 'version'
HYPERVISOR_HOSTNAME = 'OS-EXT-SRV-ATTR:hypervisor_hostname'
VM_STATE = 'OS-EXT-STS:vm_state'

# os-net-config constants
OVS_AGENT_INI_FILEPATH = '/etc/neutron/plugins/ml2/openvswitch_agent.ini'
RH_NET_CONF_PATH = "/etc/os-net-config/config.json"



# Copyright 2014 Big Switch Networks, Inc.
# All Rights Reserved.
#
# Copyright 2011 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
import time

import eventlet
import oslo_messaging

from neutron_lib.agent import topics
from neutron_lib import constants as q_const
from neutron_lib import context as q_context
from oslo_config import cfg
from oslo_log import log
from oslo_service import loopingcall

from networking_bigswitch.plugins.bigswitch import config as pl_config
from networking_bigswitch.plugins.bigswitch.i18n import _LI
from neutron.agent.common import ovs_lib
from neutron.agent import rpc as agent_rpc
from neutron.agent import securitygroups_rpc as sg_rpc
from neutron.api.rpc.handlers import securitygroups_rpc as api_sg_rpc
from neutron.common import config
from neutron.extensions import securitygroup as ext_sg

eventlet.monkey_patch()

LOG = log.getLogger(__name__)


class FilterDeviceIDMixin(sg_rpc.SecurityGroupAgentRpc):
    """Override SecurityGroupAgentRpc methods that call firewall_driver.

    This is to ensure that device ID sent to firewall driver is always without
    any prefixes. Since the firewall_driver adds the prefix regardless of
    whether it already has or not AND when reading local maps/dicts in
    firewall driver, it tries to match the start of device ID _without_ prefix,
    it fails and goes in an inconsistent state when adding and removing
    interfaces and doing firewall operations.

    Hence the name FilterDeviceIDMixin :)

    """

    def prepare_devices_filter(self, device_ids):
        if not device_ids:
            return
        LOG.info(_LI("Preparing filters for devices %r"), device_ids)
        self._apply_port_filter(device_ids)

    def _apply_port_filter(self, device_ids, update_filter=False):
        # use tap as a prefix because ml2 is hard-coded to expect that
        device_ids = [d.replace('qvo', 'tap') for d in device_ids]
        LOG.info(_LI("Applying port filter for devices %r"), device_ids)

        if self.use_enhanced_rpc:
            devices_info = self.plugin_rpc.security_group_info_for_devices(
                self.context, list(device_ids))
            devices = devices_info['devices']
            security_groups = devices_info['security_groups']
            security_group_member_ips = devices_info['sg_member_ips']
        else:
            devices = self.plugin_rpc.security_group_rules_for_devices(
                self.context, list(device_ids))

        with self.firewall.defer_apply():
            if self.use_enhanced_rpc:
                LOG.debug("Update security group information for ports %s",
                          list(devices.keys()))
                self._update_security_group_info(
                    security_groups, security_group_member_ips)
            for device in devices.values():
                # strip off 'tap' prefix from device. all calls to
                # self.firewall i.e. firewall_driver must pass plain device-id
                # without annotations. firewall_driver adds the appropriate
                # prefix
                device['device'] = device['device'].replace('tap', '')
                if update_filter:
                    LOG.debug("Update port filter for %s", device['device'])
                    self.firewall.update_port_filter(device)
                else:
                    LOG.debug("Prepare port filter for %s", device['device'])
                    self.firewall.prepare_port_filter(device)

    def refresh_firewall(self, device_ids=None):
        LOG.info(_LI("Refresh firewall rules"))
        if not device_ids:
            device_ids = self.firewall.ports.keys()
            if not device_ids:
                LOG.info(_LI("No ports here to refresh firewall"))
                return
        self._apply_port_filter(device_ids, update_filter=True)

    def remove_devices_filter(self, device_ids):
        if not device_ids:
            return
        # strip off 'qvo' prefix from device. all calls to
        # self.firewall i.e. firewall_driver must pass plain device-id
        # without annotations. firewall_driver adds the appropriate
        # prefix
        device_ids = [d.replace('qvo', '') for d in device_ids]
        LOG.info(_LI("Remove device filter for %r"), device_ids)
        with self.firewall.defer_apply():
            for device_id in device_ids:
                device = self.firewall.ports.get(device_id)
                if not device:
                    continue
                self.firewall.remove_port_filter(device)


class RestProxyAgent(api_sg_rpc.SecurityGroupAgentRpcCallbackMixin):

    target = oslo_messaging.Target(version='1.1')

    def __init__(self, integ_br, polling_interval):
        super(RestProxyAgent, self).__init__()
        self.polling_interval = polling_interval

        self.int_br = ovs_lib.OVSBridge(integ_br)
        self.agent_type = "OVS Agent"
        self.agent_state = {
            'binary': 'neutron-bsn-agent',
            'host': cfg.CONF.host,
            'topic': q_const.L2_AGENT_TOPIC,
            'configurations': {},
            'agent_type': self.agent_type,
            'start_flag': True}
        self.use_call = True

        self._setup_rpc()
        self.sg_agent = FilterDeviceIDMixin(self.context, self.sg_plugin_rpc)

    def _report_state(self):
        # How many devices are likely used by a VM
        try:
            self.state_rpc.report_state(self.context,
                                        self.agent_state,
                                        self.use_call)
            self.use_call = False
            self.agent_state.pop('start_flag', None)
        except Exception:
            LOG.exception("Failed reporting state!")

    def _setup_rpc(self):
        self.topic = topics.AGENT
        self.plugin_rpc = agent_rpc.PluginApi(topics.PLUGIN)
        self.sg_plugin_rpc = api_sg_rpc.SecurityGroupServerRpcApi(
            topics.PLUGIN)
        self.context = q_context.get_admin_context_without_session()
        self.endpoints = [self]
        consumers = [[topics.PORT, topics.UPDATE],
                     [topics.SECURITY_GROUP, topics.UPDATE]]
        self.connection = agent_rpc.create_consumers(self.endpoints,
                                                     self.topic,
                                                     consumers)
        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.PLUGIN)
        report_interval = cfg.CONF.AGENT.report_interval
        if report_interval:
            heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            heartbeat.start(interval=report_interval)

    def port_update(self, context, **kwargs):
        LOG.debug("Port update received")
        port = kwargs.get('port')
        vif_port = self.int_br.get_vif_port_by_id(port['id'])
        if not vif_port:
            LOG.debug("Port %s is not present on this host.", port['id'])
            return

        LOG.debug("Port %s found. Refreshing firewall.", port['id'])
        if ext_sg.SECURITYGROUPS in port:
            self.sg_agent.refresh_firewall()

    def _update_ports(self, registered_ports):
        ports = self.int_br.get_vif_port_set()
        if ports == registered_ports:
            return
        added = ports - registered_ports
        removed = registered_ports - ports
        return {'current': ports,
                'added': added,
                'removed': removed}

    def _process_devices_filter(self, port_info):
        if 'added' in port_info:
            self.sg_agent.prepare_devices_filter(port_info['added'])
        if 'removed' in port_info:
            self.sg_agent.remove_devices_filter(port_info['removed'])

    def daemon_loop(self):
        ports = set()

        while True:
            start = time.time()
            try:
                port_info = self._update_ports(ports)
                if port_info:
                    LOG.debug("Agent loop has new device")
                    self._process_devices_filter(port_info)
                    ports = port_info['current']
            except Exception:
                LOG.exception("Error in agent event loop")

            elapsed = max(time.time() - start, 0)
            if (elapsed < self.polling_interval):
                time.sleep(self.polling_interval - elapsed)
            else:
                LOG.debug("Loop iteration exceeded interval "
                          "(%(polling_interval)s vs. %(elapsed)s)!",
                          {'polling_interval': self.polling_interval,
                           'elapsed': elapsed})


def main():
    config.init(sys.argv[1:])
    config.setup_logging()
    pl_config.register_config()

    integ_br = cfg.CONF.RESTPROXYAGENT.integration_bridge
    polling_interval = cfg.CONF.RESTPROXYAGENT.polling_interval
    bsnagent = RestProxyAgent(integ_br, polling_interval)
    bsnagent.daemon_loop()
    sys.exit(0)

if __name__ == "__main__":
    main()
